# -*- coding: utf-8 -*-
# @Date    : 2020/6/29
# @Author  : mingming.xu
# @Email   : xv44586@gmail.com
# @File    : models.py
import six
import json
import numpy as np

from toolkit4nlp.layers import *
from keras.models import Model
from toolkit4nlp.utils import remove_arguments, string_matching, insert_arguments


class Transformer(object):
    """
    base class
    """

    def __init__(self,
                 vocab_size,  # 词表大小
                 hidden_size,  # 编码维度
                 num_hidden_layers,  # Transformer总的层数
                 num_attention_heads,  # Attention头数
                 intermediate_size,  # FeedForward 层的隐层维度
                 hidden_act=None,  # FeedFoward 层的激活函数
                 dropout_rate=None,  # dropout 比例
                 embedding_size=None,  # embedding 层维度
                 attention_key_size=None,  # attentioin 中 Q, K 的head size
                 sequence_length=None,  # 是否固定序列长度
                 layers=None,  # 外部传人的layer
                 name=None,  # 模型名
                 prefix=None,  # layer name 的前缀
                 keep_tokens=None,  # 自定义保留token
                 with_residual_attention=False,  # attention scores 残差
                 skip_weights_from_checkpoints=[],  # 从checkpoints中load weights 跳过的权重，会同时作用在trainable_weights 和 variables
                 **kwargs):
        """
        """
        if keep_tokens is not None:
            vocab_size = len(keep_tokens)

        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.num_hidden_layers = num_hidden_layers
        self.num_attention_heads = num_attention_heads
        if hidden_size % num_attention_heads != 0:
            raise ValueError('The hidden size {hidden_size} is not a multiple of the number of attention heads {'
                             'num_attention_heads}')
        self.attention_head_size = hidden_size // num_attention_heads
        self.attention_key_size = attention_key_size or self.attention_head_size
        self.intermediate_size = intermediate_size
        self.hidden_act = hidden_act
        self.dropout_rate = dropout_rate or 0
        self.embedding_size = embedding_size or hidden_size
        self.sequence_length = sequence_length
        self.position_bias = None
        self.layers = layers or {}
        self.name = name
        self.prefix = prefix or ''
        self.keep_tokens = keep_tokens
        self.with_residual_attention = with_residual_attention
        self.attention_bias = None  # 包含对attention 的偏移，如attention mask
        self.attention_scores = None  # 记录attention scores，用来实现residual attention
        self.skip_weights_from_checkpoints = skip_weights_from_checkpoints
        self.built = False

    def build(self,
              layer_norm_cond=None,
              layer_norm_cond_hidden_size=None,
              layer_norm_cond_hidden_act=None,
              additional_input_layers=None,
              **kwargs):
        """构建模型
        layer_norm_*为实现conditional layer normalization时使用。
        additional_input_layers 为新增的输入
        """
        if self.built:
            return None

        # inputs
        inputs = self.get_inputs()
        self.set_inputs(inputs, additional_input_layers)
        # conditional layer norm
        self.layer_norm_conds = [
            layer_norm_cond,
            layer_norm_cond_hidden_size,
            layer_norm_cond_hidden_act or 'linear'
        ]
        # call
        outputs = self.call(inputs)
        self.set_outputs(outputs)
        # model
        self.model = Model(self.inputs, self.outputs, name=self.name)

        self.built = True

    def call(self, inputs):
        """模型的计算图
        """
        # embedding
        outputs = self.apply_embeddings(inputs)

        # main transformer layers
        for _ in range(self.num_hidden_layers):
            outputs = self.apply_transformer_layers(outputs, _)

        # task related layers
        outputs = self.apply_task_related(outputs)
        return outputs

    def prefixed(self, name):
        """增加前缀
        """
        if name is not None:
            return self.prefix + name

    def apply(self, inputs=None, layer=None, name=None, arguments=None, **kwargs):
        """
         记录layer信息方便后续mapping权重服务；重用同名layer;
         layer(name=layer_name, **kwargs)(inputs, **arguments)
        :param inputs: 上一层的输出
        :param layer: 具体layer
        :param name: 层的名字
        :param arguments: 计算时使用参数
        :param kwargs: 初始化参数
        :return:
        """
        if layer is Dropout and self.dropout_rate == 0:
            return inputs

        arguments = {} if arguments is None else arguments

        # add prefix
        name = self.prefixed(name)
        kwargs['name'] = name

        if name not in self.layers:
            current_layer = layer(**kwargs)
            name = current_layer.name
            self.layers[name] = current_layer

        if inputs is None:
            return self.layers[name]

        return self.layers[name](inputs, **arguments)

    def get_inputs(self):
        raise NotImplementedError

    def apply_embeddings(self, inputs):
        raise NotImplementedError

    def apply_transformer_layers(self, inputs, idx):
        raise NotImplementedError

    def apply_task_related(self, inputs):
        raise NotImplementedError

    def set_inputs(self, inputs, additional_input_layers=None):
        """设置input 和 inputs
        """
        if inputs is None:
            inputs = []
        elif not isinstance(inputs, list):
            inputs = [inputs]

        inputs = inputs[:]
        if additional_input_layers is not None:
            if not isinstance(additional_input_layers, list):
                additional_input_layers = [additional_input_layers]
            inputs.extend(additional_input_layers)

        self.inputs = inputs
        if len(inputs) > 1:
            self.input = inputs
        else:
            self.input = inputs[0]

    def set_outputs(self, outputs):
        """设置output 和 outputs
        """
        if not isinstance(outputs, list):
            outputs = [outputs]

        outputs = outputs[:]
        self.outputs = outputs
        if len(outputs) > 1:
            self.output = outputs
        else:
            self.output = outputs[0]

    @property
    def initializer(self):
        """
        截断正态分布初始化
        """
        return keras.initializers.TruncatedNormal(stddev=0.02)

    def load_variable(self, checkpoint, name):
        return tf.train.load_variable(checkpoint, name)

    def load_embeddings(self, embeddings):
        if self.keep_tokens:
            return embeddings[self.keep_tokens]
        return embeddings

    def create_variable(self, name, value):
        return tf.Variable(initial_value=value, name=name)

    def variable_mapping(self):
        # keras 层与checkpoint变量直接的映射关系
        return {}

    def load_weights_from_checkpoint(self, checkpoint, mapping=None):
        """根据 mapping 从 checkpoint 加载权重
        """
        mapping = mapping or self.variable_mapping()
        mapping = {self.prefixed(k): v for k, v in mapping.items()}
        mapping = {k: v for k, v in mapping.items() if k in self.layers}

        weights_values_pairs = []
        for layer_name, variables in mapping.items():
            weights = self.layers[layer_name].trainable_weights

            # skip weights by names
            if self.skip_weights_from_checkpoints:
                weights = [w for w in weights if not string_matching(w.name, self.skip_weights_from_checkpoints)]
                variables = [v for v in values if not string_matching(v, self.skip_weights_from_checkpoints)]
            values = [self.load_variable(checkpoint, v) for v in variables]
            weights_values_pairs.extend(zip(weights, values))

        K.batch_set_value(weights_values_pairs)

    def save_weights_as_checkpoint(self, checkpoint_path, mapping=None):
        """根据mapping 将权重保存为 checkpoint格式
        """
        mapping = mapping or self.variable_mapping()
        mapping = {self.prefixed(k): v for k, v in mapping.items()}
        mapping = {k: v for k, v in mapping.items() if k in self.layers}

        with tf.Graph().as_default():
            for layer_name, variables in mapping.items():
                layer = self.layers[layer_name]
                weights = K.batch_get_value(layer.trainable_weights)
                for variable, weight in zip(variables, weights):
                    self.create_variable(variable, weight)
            with tf.Session() as sess:
                sess.run(tf.global_variables_initializer())
                saver = tf.train.Saver()
                saver.save(sess, checkpoint_path, write_meta_graph=False)

    def simplify(self, inputs):
        """过滤列表中的None"""
        inputs = [i for i in inputs if i is not None]
        if len(inputs) == 1:
            return inputs[0]
        return inputs

    def compute_attention_bias(self, inputs):
        return self.attention_bias

    def apply_attention(self, inputs, attention_name, arguments):
        # attention 层有很多变体，所以单独抽出来一个方法，来适应不同变体
        x = self.apply(inputs,
                       MultiHeadAttention,
                       name=attention_name,
                       head_nums=self.num_attention_heads,
                       head_size=self.attention_head_size,
                       arguments=arguments,
                       kernel_initializer=self.initializer,
                       with_residual_attention=self.with_residual_attention)
        if self.with_residual_attention:
            x, att_scores = x
            self.attention_scores = att_scores

        return x


class BERT(Transformer):
    """
    Google Bert
    """

    def __init__(self,
                 max_position,  # 序列最大长度
                 with_pool=False,  # 是否包含pooler部分
                 with_nsp=False,  # 是否包含NSP部分
                 with_mlm=False,  # 是否包含mlm部分
                 type_vocab_size=2,  # segment type 的种类
                 shared_segment_embeddings=False,  # segment 与 token 是否共用embedding
                 **kwargs
                 ):
        super(BERT, self).__init__(**kwargs)
        self.max_position = max_position

        self.with_pool = with_pool
        self.with_nsp = with_nsp
        self.with_mlm = with_mlm
        self.type_vocab_size = type_vocab_size
        self.shared_segment_embeddings=shared_segment_embeddings
        # nsp need pooler
        if with_nsp and not with_pool:
            self.with_pool = True

    def get_inputs(self):
        # 新增type_vocab_size，来过滤segment-inputs
        token_in = self.apply(layer=Input,
                              name='Input-Token',
                              shape=(self.sequence_length,))
        inputs = [token_in]
        if self.type_vocab_size > 0:
            segment_in = self.apply(layer=Input,
                                name='Input-Segment',
                                shape=(self.sequence_length,))
            inputs.append(segment_in)

        return inputs

    def apply_embeddings(self, inputs):
        """token_embedding + segment_embedding + position_embedding
        """
        inputs = inputs[:]
        x = inputs.pop(0)
        if self.type_vocab_size > 0:
            s = inputs.pop(0)
        # condition layer norm
        z = self.layer_norm_conds[0]

        x = self.apply(inputs=x,
                     layer=Embedding,
                     name='Embedding-Token',
                     input_dim=self.vocab_size,
                     output_dim=self.embedding_size,
                     embeddings_initializer=self.initializer,
                     mask_zero=True
                     )
        if self.type_vocab_size > 0:
            if self.shared_segment_embeddings:
                name = 'Embedding-Token'
            else:
                name = 'Embedding-Segment'
            s = self.apply(s,
                           Embedding,
                           name=name,
                           input_dim=self.type_vocab_size,
                           output_dim=self.embedding_size,
                           embeddings_initializer=self.initializer,
                                           )
            x = self.apply([x, s], Add, name='Embedding-Token-Segment')
        x = self.apply(x,
                       PositionEmbedding,
                       name='Embedding-Position',
                       input_dim=self.max_position,
                       output_dim=self.embedding_size,
                       embeddings_initializer=self.initializer,
                       merge_mode='add')

        x = self.apply(inputs=self.simplify([x, z]),
                       layer=LayerNormalization,
                       conditional=(z is not None),
                       condition_hidden_units=self.layer_norm_conds[1],
                       condition_hidden_activation=self.layer_norm_conds[2],
                       condition_hidden_initializer=self.initializer,
                       name='Embedding-Norm')
        x = self.apply(x,
                       Dropout,
                       name='Embedding-Dropout',
                       rate=self.dropout_rate)
        if self.hidden_size != self.embedding_size:
            x = self.apply(x,
                           Dense,
                           name='Embedding-Mapping',
                           units=self.hidden_size,
                           kernel_initializer=self.initializer)

        return x

    def apply_transformer_layers(self, inputs, idx):
        """
        Att --> Dropout --> Add --> LN --> FFN --> Dropout -->  Add --> LN
        """
        attention_name = 'Transformer-%d-MultiHeadSelfAttention' % idx
        feed_forward_name = 'Transformer-%d-FeedForward' % idx
        attention_bias = self.compute_attention_bias(idx)

        x_pre, x = inputs, [inputs, inputs, inputs]
        z = self.layer_norm_conds[0]
        arguments = {'a_bias': None}
        if attention_bias is not None:
            arguments['a_bias'] = True
            x.append(attention_bias)

        # self-attention
        x = self.apply_attention(x, attention_name, arguments)

        x = self.apply(x,
                       Dropout,
                       name='%s-Dropout' % attention_name,
                       rate=self.dropout_rate)

        x = self.apply([x_pre, x],
                       Add,
                       name='%s-Add' % attention_name
                       )

        x = self.apply(inputs=self.simplify([x, z]),
                       layer=LayerNormalization,
                       conditional=(z is not None),
                       condition_hidden_units=self.layer_norm_conds[1],
                       condition_hidden_activation=self.layer_norm_conds[2],
                       condition_hidden_initializer=self.initializer,
                       name='%s-Norm' % attention_name,
                       )

        # feedforward
        x_pre = x
        x = self.apply(x,
                       FeedForward,
                       name=feed_forward_name,
                       units=self.intermediate_size,
                       activation=self.hidden_act,
                       kernel_initializer=self.initializer
                       )
        x = self.apply(x,
                       Dropout,
                       name='%s-Dropout' % feed_forward_name,
                       rate=self.dropout_rate)
        x = self.apply([x_pre, x],
                       Add,
                       name='%s-Add' % feed_forward_name)
        x = self.apply(inputs=self.simplify([x, z]),
                       layer=LayerNormalization,
                       conditional=(z is not None),
                       condition_hidden_units=self.layer_norm_conds[1],
                       condition_hidden_activation=self.layer_norm_conds[2],
                       condition_hidden_initializer=self.initializer,
                       name='%s-Norm' % feed_forward_name)
        return x

    def apply_task_related(self, inputs):
        """
        跟据不同task加不同的layer产生不同的outputs
        :param inputs:
        :return:
        """
        x = inputs
        outputs = [x]
        z = self.layer_norm_conds[0]
        if self.with_pool:
            # pooler 提取cls向量
            x = self.apply(x, layer=Lambda, name='Pooler', function=lambda x: x[:, 0])
            pool_activation = 'tanh' if self.with_pool is True else self.with_pool
            x = self.apply(x,
                           layer=Dense,
                           name='Pooler-Dense',
                           units=self.hidden_size,
                           activation=pool_activation,
                           kernel_initializer=self.initializer)
            if self.with_nsp:
                # Next sentence prediction
                x = self.apply(x,
                               layer=Dense,
                               name='NSP-Proba',
                               units=2,
                               activation='softmax',
                               kernel_initializer=self.initializer)

            outputs.append(x)

        if self.with_mlm:
            # Mask language model, Dense --> Norm --> Embedding --> Biasadd --> softmax
            x = outputs[0]
            x = self.apply(x,
                           layer=Dense,
                           name='MLM-Dense',
                           units=self.embedding_size,
                           activation=self.hidden_act,
                           kernel_initializer=self.initializer)
            x = self.apply(inputs=self.simplify([x, z]),
                           layer=LayerNormalization,
                           conditional=(z is not None),
                           condition_hidden_units=self.layer_norm_conds[1],
                           condition_hidden_activation=self.layer_norm_conds[2],
                           condition_hidden_initializer=self.initializer,
                           name='MLM-Norm',
                           )
            # 重用embedding-token layer
            x = self.apply(x, Embedding, 'Embedding-Token', arguments={'mode': 'dense'})
            x = self.apply(x, BiasAdd, 'MLM-Bias')
            mlm_activation = 'softmax' if self.with_mlm is True else self.with_mlm
            x = self.apply(x, Activation, 'MLM-Activation', activation=mlm_activation)
            outputs.append(x)

        if len(outputs) == 1:
            return outputs[0]
        if len(outputs) == 2:
            return outputs[1]

        return outputs[1:]

    def load_variable(self, checkpoint, name):
        """加载单个变量的函数
        """
        variable = super(BERT, self).load_variable(checkpoint, name)
        if name == 'cls/seq_relationship/output_weights':
            return variable.T
        elif name in ['bert/embeddings/word_embeddings',
                      'cls/predictions/output_bias', ]:
            return self.load_embeddings(variable)
        else:
            return variable

    def variable_mapping(self):
        """映射到官方BERT权重格式
        """
        mapping = {
            'Embedding-Token': ['bert/embeddings/word_embeddings'],
            'Embedding-Segment': ['bert/embeddings/token_type_embeddings'],
            'Embedding-Position': ['bert/embeddings/position_embeddings'],
            'Embedding-Norm': [
                'bert/embeddings/LayerNorm/beta',
                'bert/embeddings/LayerNorm/gamma',
            ],
            'Embedding-Mapping': [
                'bert/encoder/embedding_hidden_mapping_in/kernel',
                'bert/encoder/embedding_hidden_mapping_in/bias',
            ],
            'Pooler-Dense': [
                'bert/pooler/dense/kernel',
                'bert/pooler/dense/bias',
            ],
            'NSP-Proba': [
                'cls/seq_relationship/output_weights',
                'cls/seq_relationship/output_bias',
            ],
            'MLM-Dense': [
                'cls/predictions/transform/dense/kernel',
                'cls/predictions/transform/dense/bias',
            ],
            'MLM-Norm': [
                'cls/predictions/transform/LayerNorm/beta',
                'cls/predictions/transform/LayerNorm/gamma',
            ],
            'MLM-Bias': ['cls/predictions/output_bias'],
        }

        for i in range(self.num_hidden_layers):
            prefix = 'bert/encoder/layer_%d/' % i
            mapping.update({
                'Transformer-%d-MultiHeadSelfAttention' % i: [
                    prefix + 'attention/self/query/kernel',
                    prefix + 'attention/self/query/bias',
                    prefix + 'attention/self/key/kernel',
                    prefix + 'attention/self/key/bias',
                    prefix + 'attention/self/value/kernel',
                    prefix + 'attention/self/value/bias',
                    prefix + 'attention/output/dense/kernel',
                    prefix + 'attention/output/dense/bias',
                ],
                'Transformer-%d-MultiHeadSelfAttention-Norm' % i: [
                    prefix + 'attention/output/LayerNorm/beta',
                    prefix + 'attention/output/LayerNorm/gamma',
                ],
                'Transformer-%d-FeedForward' % i: [
                    prefix + 'intermediate/dense/kernel',
                    prefix + 'intermediate/dense/bias',
                    prefix + 'output/dense/kernel',
                    prefix + 'output/dense/bias',
                ],
                'Transformer-%d-FeedForward-Norm' % i: [
                    prefix + 'output/LayerNorm/beta',
                    prefix + 'output/LayerNorm/gamma',
                ],
            })

        return mapping


class LM_Mask(object):
    """
    下三角Attention Mask，主要用于自回归语言模型
    """

    def compute_attention_bias(self, inputs=None):
        """重写attention mask 计算逻辑：全局下三角矩阵，形如：
        [[[[1, 0, 0]
        [1, 1, 0]
        [1, 1, 1]]]]
        """
        if self.attention_bias is None:
            def compute_lm_mask(segments):
                seq_len = K.shape(segments)[1]
                idx = K.arange(0, seq_len)
                mask = idx[None, :] <= idx[:, None]
                mask = K.cast(mask, K.floatx())
                return -(1 - mask[None, None]) * 1e12

            self.attention_bias = self.apply(inputs=self.inputs[0],
                                             layer=Lambda,
                                             function=compute_lm_mask,
                                             name='Attention-LM-Mask')

        return self.attention_bias


class UniLM_Mask(object):
    """
    UniLM形式的attention mask
    """

    def compute_attention_bias(self, inputs=None):
        """重写attention mask 计算逻辑,segment 1 全部为1，segment 2 中为下三角矩阵.
        假如 输入序列为 [CLS] [seg1] [SEP] [seg2] [SEP], 对应 mask 为：
        [[1，1，1, 0, 0]
        [1, 1, 1, 0, 0]
        [1, 1, 1, 0, 0]
        [1, 1, 1, 1, 0]
        [1, 1, 1, 1, 1,]]
        """
        if self.attention_bias is None:
            def compute_unilm_mask(segments):
                idx = K.cumsum(segments, axis=1)
                mask = idx[:, None, :] <= idx[:, :, None]
                mask = K.cast(mask, K.floatx())
                return -(1 - mask[:, None]) * 1e12

            self.attention_bias = self.apply(inputs=self.inputs[1],
                                             layer=Lambda,
                                             function=compute_unilm_mask,
                                             name='Attention-UniLM-Attention')

        return self.attention_bias


class GPT(LM_Mask, BERT):
    """
    OPENAI GPT
    """
    @insert_arguments(final_activation='softmax')
    @remove_arguments('with_pool', 'with_mlm', 'with_nsp')
    def __init__(self, **kwargs):
        super(GPT, self).__init__(**kwargs)

    def apply_embeddings(self, inputs):
        """token_embedding + segment_embedding + position_embedding, 其中 segment_embedding 共用 token_embedding
        与BERT 的主要区别是没有LayerNormalization
        """
        inputs = inputs[:]
        x = inputs.pop(0)
        if self.type_vocab_size > 0:
            s = inputs.pop(0)

        x = self.apply(inputs=x,
                     layer=Embedding,
                     name='Embedding-Token',
                     input_dim=self.vocab_size,
                     output_dim=self.embedding_size,
                     embeddings_initializer=self.initializer,
                     mask_zero=True
                     )

        if self.type_vocab_size > 0:
            if self.shared_segment_embeddings:
                name = 'Embedding-Token'
            else:
                name = 'Embedding-Segment'

            s = self.apply(s,
                           Embedding,
                           name=name,
                           input_dim=self.type_vocab_size,
                           output_dim=self.embedding_size,
                           embeddings_initializer=self.initializer,
                           )
            x = self.apply([x, s], Add, name='Embedding-Token-Segment')
        x = self.apply(x,
                       PositionEmbedding,
                       name='Embedding-Position',
                       input_dim=self.max_position,
                       output_dim=self.embedding_size,
                       embeddings_initializer=self.initializer,
                       merge_mode='add')

        x = self.apply(x,
                       Dropout,
                       name='Embedding-Dropout',
                       rate=self.dropout_rate)
        if self.hidden_size != self.embedding_size:
            x = self.apply(x,
                           Dense,
                           name='Embedding-Mapping',
                           units=self.hidden_size,
                           kernel_initializer=self.initializer)

        return x

    def apply_task_related(self, inputs):
        """
        Language Model部分
        """
        x = inputs
        x = self.apply(inputs=x,
                       layer=Embedding,
                       arguments={'mode': 'dense'},
                       name='Embedding-Token')
        x = self.apply(inputs=x,
                       layer=Activation,
                       activation=self.final_activation,
                       name='LM-Actication'
                       )
        return x

    def load_variable(self, checkpoint, name):
        variable = super(GPT, self).load_variable(checkpoint, name)
        if name == 'gpt/embeddings/word_embeddings':
            return self.load_embeddings(variable)
        else:
            return variable

    def variable_mapping(self):
        mapping = super(GPT, self).variable_mapping()
        mapping = {
            k: [i.replace('bert/','gpt/').replace('encoder', 'transformer') for i in v] for k, v in mapping.items()
        }
        return mapping


class GPT2(GPT):
    """
    GPT2
    """

    def get_inputs(self):
        x = self.apply(layer=Input, shape=(self.sequence_length,), name='Input-Token')
        return x

    def apply_embeddings(self, inputs):
        """
        GPT2的embedding 是token_embedding + position_embedding
        """
        x = inputs

        x = self.apply(inputs=x,
                       layer=Embedding,
                       input_dim=self.vocab_size,
                       output_dim=self.embedding_size,
                       embeddings_initializer=self.initializer,
                       mask_zero=True,
                       name='Embedding-Token'
                       )
        x = self.apply(inputs=x,
                       layer=PositionEmbedding,
                       input_dim=self.max_position,
                       output_dim=self.embedding_size,
                       merge_mode='add',
                       embeddings_initializer=self.initializer,
                       name='Embedding-Position')
        if self.embedding_size != self.hidden_size:
            x = self.apply(inputs=x,
                           layer=Dense,
                           units=self.hidden_size,
                           kernel_initializer=self.initializer,
                           name='Embedding-Mapping')

        return x

    def apply_transformer_layers(self, inputs, idx):
        """
        LN --> Att --> Dropout --> Add --> LN --> FFN --> Dropout --> Add
        """
        x = inputs
        z = self.layer_norm_conds[0]

        attention_name = 'Transformer-%d-MultiHeadSelfAttention' % idx
        feed_forward_name = 'Transformer-%d-FeedForward' % idx

        attention_bias = self.compute_attention_bias(idx)

        # self-attention
        x_pre = x

        x = self.apply(inputs=self.simplify([x, z]),
                       layer=LayerNormalization,
                       epsilon=1e-5,
                       conditional=(z is not None),
                       condition_hidden_units=self.layer_norm_conds[1],
                       condition_hidden_activation=self.layer_norm_conds[2],
                       condition_hidden_initializer=self.initializer,
                       name='%s-Norm' % attention_name,
                       )

        x = self.apply_attention([x, x, x, attention_bias], attention_name, arguments={'a_bias': True})

        x = self.apply(x,
                       Dropout,
                       name='%s-Dropout' % attention_name,
                       rate=self.dropout_rate)

        x = self.apply([x_pre, x],
                       Add,
                       name='%s-Add' % attention_name
                       )

        # feedforward
        x_pre = x
        x = self.apply(inputs=self.simplify([x, z]),
                       layer=LayerNormalization,
                       epsilon=1e-5,
                       conditional=(z is not None),
                       condition_hidden_units=self.layer_norm_conds[1],
                       condition_hidden_activation=self.layer_norm_conds[2],
                       condition_hidden_initializer=self.initializer,
                       name='%s-Norm' % feed_forward_name)
        x = self.apply(x,
                       FeedForward,
                       name=feed_forward_name,
                       units=self.intermediate_size,
                       activation=self.hidden_act,
                       kernel_initializer=self.initializer
                       )
        x = self.apply(x,
                       Dropout,
                       name='%s-Dropout' % feed_forward_name,
                       rate=self.dropout_rate)
        x = self.apply([x_pre, x],
                       Add,
                       name='%s-Add' % feed_forward_name)

        return x

    def apply_task_related(self, inputs):
        """GPT2做language model时先做了LanyerNormalization + Dropout，然后预测token
        """
        x = inputs
        z = self.layer_norm_conds[0]

        x = self.apply(inputs=self.simplify([x, z]),
                       layer=LayerNormalization,
                       epsilon=1e-5,
                       conditional=(z is not None),
                       condition_hidden_units=self.layer_norm_conds[1],
                       condition_hidden_activation=self.layer_norm_conds[2],
                       condition_hidden_initializer=self.initializer,
                       name='Output-Norm')
        x = self.apply(
            inputs=x,
            layer=Dropout,
            rate=self.dropout_rate,
            name='Output-Dropout'
        )
        x = super(GPT2, self).apply_task_related(x)
        return x

    def variable_mapping(self):
        mapping = super(GPT2, self).variable_mapping()
        mapping = {
            k: [i.replace('output/LayerNorm', 'input/LayerNorm') for i in v]
            for k, v in mapping.items()
        }
        mapping['Output-Norm'] = ['gpt/output/LayerNorm/beta', 'gpt/output/LayerNorm/gamma']
        return mapping


class GPT2ForSequenceClassification(GPT2):
    """
    GPT2 做分类任务，模型的改变主要是将lm_head 去掉拼了一个dense 层，同时只使用了最后一个'有效' token 对应的output
    当label ==1 时，做回归任务，当label > 1 时，为分类任务
    """
    def __init__(self, num_labels=1, pad_token_id=None, **kwargs):
        super(GPT2ForSequenceClassification, self).__init__(**kwargs)
        self.num_labels = num_labels
        self.pad_token_id = pad_token_id  # 用来寻找最后一个有效token 的位置，如果不指定，则选最后一个
        self.final_activation = 'sigmoid' if num_labels == 1 else 'softmax'

    def apply_task_related(self, inputs):
        """LanyerNormalization --> Dropout  --> GatherLastToken  --> Dense
        """
        x = inputs
        z = self.layer_norm_conds[0]

        x = self.apply(inputs=self.simplify([x, z]),
                       layer=LayerNormalization,
                       epsilon=1e-5,
                       conditional=(z is not None),
                       condition_hidden_units=self.layer_norm_conds[1],
                       condition_hidden_activation=self.layer_norm_conds[2],
                       condition_hidden_initializer=self.initializer,
                       name='Output-Norm')

        x = self.apply(
            inputs=x,
            layer=Dropout,
            rate=self.dropout_rate,
            name='Output-Dropout'
        )
        # extract last token output
        x = self.apply(
            inputs=[x, self.inputs[0]],
            layer=GatherLastToken,
            pad_token_id=self.pad_token_id,
        )

        x = self.apply(
            inputs=x,
            layer=Dense,
            units=self.num_labels,
            kernel_initializer=self.initializer,
            use_bias=False,
            name='Output-Score',
            activation=self.final_activation
        )

        return x

    def variable_mapping(self):
        mapping = super(GPT2ForSequenceClassification, self).variable_mapping()
        mapping['Output-Score'] = ['gpt/output/score']
        return mapping


class ELECTRA(BERT):
    @remove_arguments('with_mlm', 'with_pool')
    def __init__(self, max_position, **kwargs):
        super(ELECTRA, self).__init__(max_position, **kwargs)

    def variable_mapping(self):
        mapping = super(ELECTRA, self).variable_mapping()
        mapping = {k: [name.replace('bert/', 'electra/') for name in v] for k, v in mapping.items()}
        # embedding mapping
        mapping['Embedding-Mapping'] = [
            'electra/embeddings_project/kernel',
            'electra/embeddings_project/bias',
        ]
        return mapping

    def apply_task_related(self, inputs):
        return inputs


class NEZHA(BERT):
    """
    出自华为诺亚方舟实验室的NEZHA模型，结构上的改进是替换bert中的绝对位置编码为相对位置编码

    ref: [NEZHA: Neural Contextualized Representation for Chinese Language Understanding](http://arxiv.org/abs/1909.00204)
    """

    def __init__(self,
                 external_embedding_size=None,
                 external_embedding_weights=None,
                 **kwargs):
        """
        增加引入外部embedding
        """

        super(NEZHA, self).__init__(**kwargs)
        self.external_embedding_size = external_embedding_size  # 外部embedding 输出维度
        self.external_embedding_weights = external_embedding_weights  # 外部embedding 初始化权重

    def apply_embeddings(self, inputs):
        """
        embedding 是 token embedding 与 segment embedding 的和
        """
        x, s = inputs[:2]
        z = self.layer_norm_conds[0]
        token_embedding = self.apply(inputs=x,
                                     layer=Embedding,
                                     name='Embedding-Token',
                                     input_dim=self.vocab_size,
                                     output_dim=self.embedding_size,
                                     embeddings_initializer=self.initializer,
                                     mask_zero=True
                                     )
        segment_embedding = self.apply(s,
                                       Embedding,
                                       name='Embedding-Segment',
                                       input_dim=2,
                                       output_dim=self.embedding_size,
                                       embeddings_initializer=self.initializer,
                                       )
        token_with_segment = self.apply([token_embedding, segment_embedding], Add, name='Embedding-Token-Segment')

        # add External knowledge
        if self.external_embedding_size:
            external_embedding = self.apply(x,
                                            Embedding,
                                            name='Embedding-External',
                                            input_dim=self.vocab_size,
                                            output_dim=self.external_embedding_size,
                                            weights=[self.external_embedding_weights]
                                            )
            if self.external_embedding_size != self.embedding_size:
                external_embedding = self.apply(external_embedding,
                                                Dense,
                                                name='External-Embedding-Mapping',
                                                units=self.embedding_size,
                                                kernel_initializer=self.initializer)
            x = self.apply([token_with_segment, external_embedding], Add, name='Embedding-Token-Segment-External')
        else:
            x = token_with_segment

        x = self.apply(inputs=self.simplify([x, z]),
                       layer=LayerNormalization,
                       conditional=(z is not None),
                       condition_hidden_units=self.layer_norm_conds[1],
                       condition_hidden_activation=self.layer_norm_conds[2],
                       condition_hidden_initializer=self.initializer,
                       name='Embedding-Norm')
        x = self.apply(x,
                       Dropout,
                       name='Embedding-Dropout',
                       rate=self.dropout_rate)
        if self.hidden_size != self.embedding_size:
            x = self.apply(x,
                           Dense,
                           name='Embedding-Mapping',
                           units=self.hidden_size,
                           kernel_initializer=self.initializer)

        return x

    def apply_transformer_layers(self, inputs, idx):
        """
        Att --> Dropout --> Add --> LN --> FFN --> Dropout -->  Add --> LN
        """
        attention_name = 'Transformer-%d-MultiHeadSelfAttention' % idx
        feed_forward_name = 'Transformer-%d-FeedForward' % idx
        attention_bias = self.compute_attention_bias(idx)
        position_bias = self.compute_position_bias(inputs)

        x_pre, x = inputs, [inputs, inputs, inputs, position_bias]  # 加入相对位置编码
        z = self.layer_norm_conds[0]
        arguments = {'a_bias': None, 'position_bias': 'relative'}
        if attention_bias is not None:
            arguments['a_bias'] = True
            x.insert(3, attention_bias)

        x = self.apply_attention(x, attention_name, arguments)

        x = self.apply(x,
                       Dropout,
                       name='%s-Dropout' % attention_name,
                       rate=self.dropout_rate)

        x = self.apply([x_pre, x],
                       Add,
                       name='%s-Add' % attention_name
                       )

        x = self.apply(inputs=self.simplify([x, z]),
                       layer=LayerNormalization,
                       conditional=(z is not None),
                       condition_hidden_units=self.layer_norm_conds[1],
                       condition_hidden_activation=self.layer_norm_conds[2],
                       condition_hidden_initializer=self.initializer,
                       name='%s-Norm' % attention_name,
                       )

        # feedforward
        x_pre = x
        x = self.apply(x,
                       FeedForward,
                       name=feed_forward_name,
                       units=self.intermediate_size,
                       activation=self.hidden_act,
                       kernel_initializer=self.initializer
                       )
        x = self.apply(x,
                       Dropout,
                       name='%s-Dropout' % feed_forward_name,
                       rate=self.dropout_rate)
        x = self.apply([x_pre, x],
                       Add,
                       name='%s-Add' % feed_forward_name)
        x = self.apply(inputs=self.simplify([x, z]),
                       layer=LayerNormalization,
                       conditional=(z is not None),
                       condition_hidden_units=self.layer_norm_conds[1],
                       condition_hidden_activation=self.layer_norm_conds[2],
                       condition_hidden_initializer=self.initializer,
                       name='%s-Norm' % feed_forward_name)
        return x

    def compute_position_bias(self, inputs):
        """计算相对位置编码"""
        if self.position_bias is None:
            x = inputs
            self.position_bias = self.apply(
                inputs=[x, x],
                layer=RelativePositionEmbedding,
                name='Relative-Position-Embedding',
                input_dim=2 * 64 + 1,
                output_dim=self.attention_key_size,
                embedding_initializer='sinusoidal',
                trainable=False
            )

        return self.position_bias


class DBERT(BERT):
    """
    用连续的dgcnn来替换原始bert中的transformer block，来尝试压缩bert
    """

    def __init__(self, inner_hidden_size=None, **kwargs):
        prefix = kwargs.get('prefix', 'DBERT-')
        kwargs['prefix'] = prefix
        super(DBERT, self).__init__(**kwargs)
        self.inner_hidden_size = inner_hidden_size or self.attention_head_size  # dgcnn hidden dim

    def apply_transformer_layers(self, inputs, idx):
        """
            liner_in(inner_hid_dim) -> dgcnn(1) -> dgcnn(3) -> dgcnn(5) -> dgcnn(8) -> liner_out(hid_dim)
        """
        block_name = 'DBlock-%d-MultiDGCNN' % idx
        liner_in_name = 'DBlock-%d-MultiDGCNN-LinerIn' % idx
        liner_out_name = 'DBlock-%d-MultiDGCNN-LinerOut' % idx
        attention_mask = self.compute_attention_mask(idx)
        x = inputs
        x = self.apply(x,
                       Dense,
                       name=liner_in_name,
                       units=self.inner_hidden_size,
                       use_bias=False
                       )
        arguments = {}
        if attention_mask is not None:
            x.append(attention_mask)

        # dgcnn-block
        x = self.apply(x,
                       DGCNN,
                       name='{block_name}-DilationRate-{r}'.format(block_name=block_name, r=1),
                       dilation_rate=1,
                       dropout_rate=self.dropout_rate,
                       arguments=arguments)

        x = self.apply(x,
                       DGCNN,
                       name='{block_name}-DilationRate-{r}'.format(block_name=block_name, r=3),
                       dilation_rate=3,
                       dropout_rate=self.dropout_rate,
                       arguments=arguments)
        x = self.apply(x,
                       DGCNN,
                       name='{block_name}-DilationRate-{r}'.format(block_name=block_name, r=5),
                       dilation_rate=5,
                       dropout_rate=self.dropout_rate,
                       arguments=arguments)
        x = self.apply(x,
                       DGCNN,
                       name='{block_name}-DilationRate-{r}'.format(block_name=block_name, r=8),
                       dilation_rate=8,
                       dropout_rate=self.dropout_rate,
                       arguments=arguments)

        x = self.apply(x,
                       Dense,
                       name=liner_out_name,
                       units=self.hidden_size,
                       use_bias=False
                       )
        x = self.apply(x, LayerNormalization, name='DBlock-%d-MultiDGCNN-Norm' % idx)
        return x


def extend_with_language_model(BaseModel):
    """
    增加下三角mask矩阵，作为语言模型使用
    """

    class LanguageModel(LM_Mask, BaseModel):
        def __init__(self, *args, **kwargs):
            super(LanguageModel, self).__init__(*args, **kwargs)
            self.with_mlm = self.with_mlm or True  # mlm output

    return LanguageModel


def extend_with_unilm(BaseModel):
    """添加UniLM mask"""

    class UniLM(UniLM_Mask, BaseModel):
        """UnilM-V1: https://arxiv.org/pdf/1905.03197.pdf"""

        def __init__(self, *args, **kwargs):
            super(UniLM, self).__init__(*args, **kwargs)
            self.with_mlm = self.with_mlm or True  # mlm output

    return UniLM


def extend_with_residual_attention(BaseModel):
    """添加当前attention scores 与前一层attention scores的直通路"""

    class RealFormer(BaseModel):
        """RealFormer: https://arxiv.org/pdf/2012.11747.pdf"""

        def __init__(self, *args, **kwargs):
            super(RealFormer, self).__init__(*args, **kwargs)
            self.with_residual_attention = True

        def compute_attention_bias(self, inputs=None):
            """
            将residual attention scores 加入 attention bias
            :param inputs: transformer layer idx
            """
            att_bias = super(RealFormer, self).compute_attention_bias(inputs)

            if self.attention_scores is not None and att_bias is not None:
                att_bias = self.apply([self.attention_scores, att_bias],
                                      Add,
                                      name='Transformer-%d-Attention-bias-Add' % inputs,
                                      )

            att_bias = att_bias if att_bias is not None else self.attention_scores
            return att_bias

    return RealFormer


class ReZero(BERT):
    def __init__(self,
                 use_layernorm=None,  # None, pre, post, when None, then rezero
                 init_reweight=1.,  # init of reweight
                 reweight_trainable=True,
                 **kwargs,
                 ):
        super(ReZero, self).__init__(**kwargs)
        assert use_layernorm in [None, 'pre', 'post']
        self.use_layernorm = use_layernorm
        self.init_reweight = init_reweight
        self.reweight_trainable = reweight_trainable

    def apply_embeddings(self, inputs):
        """token_embedding + segment_embedding + position_embedding
        """
        x, s = inputs[:2]
        # condition layer norm
        z = self.layer_norm_conds[0]

        token_embedding = self.apply(inputs=x,
                                     layer=Embedding,
                                     name='Embedding-Token',
                                     input_dim=self.vocab_size,
                                     output_dim=self.embedding_size,
                                     embeddings_initializer=self.initializer,
                                     mask_zero=True
                                     )
        segment_embedding = self.apply(s,
                                       Embedding,
                                       name='Embedding-Segment',
                                       input_dim=self.type_vocab_size,
                                       output_dim=self.embedding_size,
                                       embeddings_initializer=self.initializer,
                                       )
        token_with_seg = self.apply([token_embedding, segment_embedding], Add, name='Embedding-Token-Segment')
        x = self.apply(token_with_seg,
                       PositionEmbedding,
                       name='Embedding-Position',
                       input_dim=self.max_position,
                       output_dim=self.embedding_size,
                       embeddings_initializer=self.initializer,
                       merge_mode='add')

        # if pre layernorm, then delete layernorm in this block
        if self.use_layernorm != 'pre':
            x = self.apply(inputs=self.simplify([x, z]),
                           layer=LayerNormalization,
                           conditional=(z is not None),
                           condition_hidden_units=self.layer_norm_conds[1],
                           condition_hidden_activation=self.layer_norm_conds[2],
                           condition_hidden_initializer=self.initializer,
                           name='Embedding-Norm')
        x = self.apply(x,
                       Dropout,
                       name='Embedding-Dropout',
                       rate=self.dropout_rate)
        if self.hidden_size != self.embedding_size:
            x = self.apply(x,
                           Dense,
                           name='Embedding-Mapping',
                           units=self.hidden_size,
                           kernel_initializer=self.initializer)

        return x

    def apply_transformer_layers(self, inputs, idx):
        """
        post: Att --> Dropout --> Add --> LN --> FFN --> Dropout -->  Add --> LN
        pre: LN --> Att --> Dropout --> Add --> LN --> FFN --> Dropout --> Add
        rezero: Att --> ReWeight --> Dropout --> Add --> FFN -->ReWeight --> Dropout --> Add
        """
        attention_name = 'Transformer-%d-MultiHeadSelfAttention' % idx
        feed_forward_name = 'Transformer-%d-FeedForward' % idx
        attention_bias = self.compute_attention_bias(idx)

        x_pre, x = inputs, inputs
        z = self.layer_norm_conds[0]
        arguments = {'a_bias': None}
        if attention_bias is not None:
            arguments['a_bias'] = True
            x.append(attention_bias)

        if self.use_layernorm == 'pre':
            x = self.apply(inputs=self.simplify([x, z]),
                           layer=LayerNormalization,
                           conditional=(z is not None),
                           condition_hidden_units=self.layer_norm_conds[1],
                           condition_hidden_activation=self.layer_norm_conds[2],
                           condition_hidden_initializer=self.initializer,
                           name='%s-Norm' % attention_name,
                           )

        x = [x, x, x]
        # self-attention
        x = self.apply_attention(x, attention_name, arguments)

        # reweight residual-connection
        x = self.apply(x,
                       ReWeight,
                       name='%s-ReWeight' % attention_name,
                       init_reweight=self.init_reweight,
                       trainable=self.reweight_trainable
                       )

        x = self.apply(x,
                       Dropout,
                       name='%s-Dropout' % attention_name,
                       rate=self.dropout_rate)

        x = self.apply([x_pre, x],
                       Add,
                       name='%s-Add' % attention_name
                       )
        if self.use_layernorm == 'post':
            x = self.apply(inputs=self.simplify([x, z]),
                           layer=LayerNormalization,
                           conditional=(z is not None),
                           condition_hidden_units=self.layer_norm_conds[1],
                           condition_hidden_activation=self.layer_norm_conds[2],
                           condition_hidden_initializer=self.initializer,
                           name='%s-Norm' % attention_name,
                           )

        # feedforward
        x_pre = x
        if self.use_layernorm == 'pre':
            x = self.apply(inputs=self.simplify([x, z]),
                           layer=LayerNormalization,
                           conditional=(z is not None),
                           condition_hidden_units=self.layer_norm_conds[1],
                           condition_hidden_activation=self.layer_norm_conds[2],
                           condition_hidden_initializer=self.initializer,
                           name='%s-Norm' % feed_forward_name
                           )

        x = self.apply(x,
                       FeedForward,
                       name=feed_forward_name,
                       units=self.intermediate_size,
                       activation=self.hidden_act,
                       kernel_initializer=self.initializer
                       )

        # reweight residual-connection
        x = self.apply(x,
                       ReWeight,
                       name='%s-ReWeight' % feed_forward_name,
                       init_reweight=self.init_reweight,
                       trainable=self.reweight_trainable
                       )
        x = self.apply(x,
                       Dropout,
                       name='%s-Dropout' % feed_forward_name,
                       rate=self.dropout_rate)

        x = self.apply([x_pre, x],
                       Add,
                       name='%s-Add' % feed_forward_name)
        if self.use_layernorm == 'post':
            x = self.apply(inputs=self.simplify([x, z]),
                           layer=LayerNormalization,
                           conditional=(z is not None),
                           condition_hidden_units=self.layer_norm_conds[1],
                           condition_hidden_activation=self.layer_norm_conds[2],
                           condition_hidden_initializer=self.initializer,
                           name='%s-Norm' % feed_forward_name)

        return x


class ALBERT(BERT):
    def apply_transformer_layers(self, inputs, idx):
        """
        Att -->  Dropout --> Add --> LN --> FFN --> Dropout --> Add --> LN
        """
        attention_name = 'Transformer-MultiHeadSelfAttention'
        feed_forward_name = 'Transformer-FeedForward'
        attention_bias = self.compute_attention_bias(idx)

        x_pre, x = inputs, [inputs, inputs, inputs]
        z = self.layer_norm_conds[0]
        arguments = {'a_bias': None}
        if attention_bias is not None:
            arguments['a_bias'] = True
            x.append(attention_bias)

        x = self.apply_attention(x, attention_name, arguments)

        x = self.apply(x,
                       Dropout,
                       name='%s-Dropout' % attention_name,
                       rate=self.dropout_rate)

        x = self.apply([x_pre, x],
                       Add,
                       name='%s-Add' % attention_name
                       )

        x = self.apply(inputs=self.simplify([x, z]),
                       layer=LayerNormalization,
                       conditional=(z is not None),
                       condition_hidden_units=self.layer_norm_conds[1],
                       condition_hidden_activation=self.layer_norm_conds[2],
                       condition_hidden_initializer=self.initializer,
                       name='%s-Norm' % attention_name,
                       )

        # feedforward
        x_pre = x
        x = self.apply(x,
                       FeedForward,
                       name=feed_forward_name,
                       units=self.intermediate_size,
                       activation=self.hidden_act,
                       kernel_initializer=self.initializer
                       )
        x = self.apply(x,
                       Dropout,
                       name='%s-Dropout' % feed_forward_name,
                       rate=self.dropout_rate)
        x = self.apply([x_pre, x],
                       Add,
                       name='%s-Add' % feed_forward_name)
        x = self.apply(inputs=self.simplify([x, z]),
                       layer=LayerNormalization,
                       conditional=(z is not None),
                       condition_hidden_units=self.layer_norm_conds[1],
                       condition_hidden_activation=self.layer_norm_conds[2],
                       condition_hidden_initializer=self.initializer,
                       name='%s-Norm' % feed_forward_name)
        return x

    def variable_mapping(self):
        mapping = super(ALBERT, self).variable_mapping()

        prefix = 'bert/encoder/transformer/group_0/inner_group_0/'
        mapping.update({
            'Transformer-MultiHeadSelfAttention': [
                prefix + 'attention_1/self/query/kernel',
                prefix + 'attention_1/self/query/bias',
                prefix + 'attention_1/self/key/kernel',
                prefix + 'attention_1/self/key/bias',
                prefix + 'attention_1/self/value/kernel',
                prefix + 'attention_1/self/value/bias',
                prefix + 'attention_1/output/dense/kernel',
                prefix + 'attention_1/output/dense/bias',
            ],
            'Transformer-MultiHeadSelfAttention-Norm': [
                prefix + 'LayerNorm/beta',
                prefix + 'LayerNorm/gamma',
            ],
            'Transformer-FeedForward': [
                prefix + 'ffn_1/intermediate/dense/kernel',
                prefix + 'ffn_1/intermediate/dense/bias',
                prefix + 'ffn_1/intermediate/output/dense/kernel',
                prefix + 'ffn_1/intermediate/output/dense/bias',
            ],
            'Transformer-FeedForward-Norm': [
                prefix + 'LayerNorm_1/beta',
                prefix + 'LayerNorm_1/gamma',
            ],
        })

        return mapping


def build_transformer_model(
        config_path=None,
        checkpoint_path=None,
        model='bert',
        application='encoder',
        return_keras_model=True,
        **kwargs
):
    """根据配置文件构建模型，可选加载checkpoint权重
    """
    configs = {}
    if config_path is not None:
        configs.update(json.load(open(config_path)))
    configs.update(kwargs)
    # convert config name
    if 'max_position' not in configs:
        configs['max_position'] = configs.get('max_position_embeddings')
    if 'dropout_rate' not in configs:
        configs['dropout_rate'] = configs.get('hidden_dropout_prob')

    models = {
        'bert': BERT,
        'reberta': BERT,
        'nezha': NEZHA,
        'electra': ELECTRA,
        'albert': ALBERT,
        'dbert': DBERT,
        'rezero': ReZero,
        'gpt': GPT,
        'gpt2': GPT2,
    }

    if isinstance(model, six.string_types):
        model = model.lower()
        MODEL = models[model]
    else:
        MODEL = model

    application = application.lower()
    if application == 'lm':
        MODEL = extend_with_language_model(MODEL)
    elif application == 'unilm':
        MODEL = extend_with_unilm(MODEL)

    if kwargs.get('with_residual_attention'):
        MODEL = extend_with_residual_attention(MODEL)

    transformer = MODEL(**configs)
    transformer.build(**configs)

    if checkpoint_path is not None:
        transformer.load_weights_from_checkpoint(checkpoint_path)

    if return_keras_model:
        return transformer.model
    else:
        return transformer



# -*- coding: utf-8 -*-
# @Date    : 2020/6/29
# @Author  : mingming.xu
# @Email   : xv44586@gmail.com
# @File    : utils.py
import re
import numpy as np
from abc import abstractmethod


def softmax(x, axis=-1):
    """numpy版softmax
    """
    x = x - x.max(axis=axis, keepdims=True)
    x = np.exp(x)
    return x / x.sum(axis=axis, keepdims=True)


class AutoRegressiveDecoder(object):
    '''
    自回归生成解码器，beam search and random sample两种策略
    '''

    def __init__(self, start_id, end_id, maxlen, minlen=None):
        self.start_id = start_id
        self.end_id = end_id
        self.maxlen = maxlen
        self.minlen = minlen or 1
        if start_id is None:
            self.first_output_ids = np.empty((1, 0), dtype=int)
        else:
            self.first_output_ids = np.array([[self.start_id]])

    def predict(self, inputs, output_ids, states, rtype='probas'):
        '''

        :param inputs:
        :param output_ids:
        :param states:
        :param rtype:logits或probas，用户定义的时候，应当根据rtype来返回不同的结果，
        rtype=probas时返回归一化的概率，rtype=logits时则返回softmax前的结果或者概率对数。
        :return: (scores, states)
        '''
        raise NotImplementedError

    def beam_search(self, inputs, beam_size, states=None, min_ends=1):
        '''

        :param inputs: [ ( seq_length,), ... ]
        :param beam_size:
        :param states:
        :param min_ends: 解码序列中 ends_token_ids 最少次数
        :return: 最优序列
        '''

        inputs = [np.array([i]) for i in inputs]

        output_ids, output_scores = self.first_output_ids, np.zeros(1)

        for step in range(self.maxlen):
            scores, states = self.predict(inputs, output_ids, states, 'logits')
            if step == 0:
                inputs = [np.repeat(i, beam_size, axis=0) for i in inputs]

            scores = output_scores.reshape((-1, 1)) + scores  # 累计得分

            indices = scores.argpartition(-beam_size, axis=None)[
                      -beam_size:]  # flatten array 然后取全局beam_size个最大score的indices
            indices_row = indices // scores.shape[1]  # 行索引， 即对应的路径索引
            indices_col = (indices % scores.shape[1]).reshape((-1, 1))  # 列索引，即token_index
            output_ids = np.concatenate([output_ids[indices_row], indices_col], axis=1)  # 将最大的token_ids 拼到对应路径上
            output_scores = np.take_along_axis(scores, indices, axis=None)  # 更新得分

            ends_counts = (output_ids == self.end_id).sum(1)  # 统计每个路径上的ends_token次数
            if output_ids.shape[1] >= self.minlen:  # 判断路径长度是否达到最短要求
                best_path_idx = output_scores.argmax()  # 得分最高路径
                if ends_counts[best_path_idx] == min_ends:  # 达到最少ends_token要求
                    return output_ids[best_path_idx]  # 返回最优路径
                else:  # 剔除已经结束但是得分不是最高的路径
                    flag = ends_counts < min_ends
                    if not flag.all():
                        inputs = [i[flag] for i in inputs]  # 删除inputs对应路径
                        output_ids = output_ids[flag]  # 删除output对应路径
                        beam_size = flag.sum()  # 重新计算beamsize
                        output_scores = output_scores[flag]
        # 达到长度后直接输出
        return output_ids[output_scores.argmax()]

    @staticmethod
    def wraps(default_rtype='probas', use_states=False):
        """用来进一步完善predict函数
        目前包含：1. 设置rtype参数，并做相应处理；
                  2. 确定states的使用，并做相应处理。
        """

        def actual_decorator(predict):
            def new_predict(
                    self, inputs, output_ids, states, rtype=default_rtype, temperatre=1
            ):
                assert rtype in ['probas', 'logits']
                prediction = predict(self, inputs, output_ids, states)

                if not use_states:
                    prediction = (prediction, None)

                if default_rtype == 'logits':
                    prediction = (
                        softmax(prediction[0] / temperatre), prediction[1]
                    )
                elif temperatre != 1:
                    probas = np.power(prediction[0], 1./temperatre)
                    probas = probas / probas.sum(axis=-1, keepdims=True)
                    prediction = (probas, prediction[1])

                if rtype == 'probas':
                    return prediction
                else:
                    return np.log(prediction[0] + 1e-12), prediction[1]

            return new_predict

        return actual_decorator

    def random_sample(self, inputs, n, topk=None, topp=None, states=None, min_ends=1):
        '''
        随机采样生成n个序列
        :param inputs:
        :param n:
        :param topk: 非None 则从概率最高的K个样本中采样
        :param topp:  非None，则从概率逆序排列后累计和不高与topp的token中采样(至少保留一个token供采样）
        :param states:
        :param min_ends: ends token出现的最少次数
        :return:
        '''
        inputs = [np.array([i]) for i in inputs]
        output_ids = self.first_output_ids
        result = []
        for step in range(self.maxlen):
            probas, states = self.predict(inputs, output_ids, states, 'probas')
            #  第一步时复制n份
            if step == 0:
                probas = np.repeat(probas, n, axis=0)
                inputs = [np.repeat(i, n, axis=0) for i in inputs]
                output_ids = np.repeat(output_ids, n, axis=0)
            if topk:
                indices_k = probas.argpartition(-topk, axis=1)[:, -topk:]  # 概率最高的K个索引
                probas = np.take_along_axis(probas, indices_k, axis=1)  # 概率最高K个
                probas /= probas.sum(1, keepdims=True)  # 重新归一化概率

            if topp:
                indices_p = probas.argsort(axis=1)[:, ::-1]  # 逆序排列
                probas = np.take_along_axis(probas, indices_p, axis=1)  # 概率逆序排列
                # 概率累计，将大于topp的置零。保证至少一个不为0
                cumsum = np.cumsum(probas, axis=1)
                #                 flag = np.roll(cumsum>=topp, 1, axis=1)
                flag = cumsum >= topp
                flag[:, 0] = False  # 后移一位并将第一位保留，以保证至少一个不为0
                probas[flag] = 0.
                probas /= probas.sum(1, keepdims=True)
            func_sample = lambda p: np.random.choice(len(p), p=p)  # 以概率p随机取一个样本
            sample_ids = np.apply_along_axis(func_sample, axis=1, arr=probas)
            sample_ids = sample_ids.reshape((-1, 1))  # 对齐output_ids
            if topp:
                sample_ids = np.take_along_axis(indices_p, sample_ids, axis=1)
            if topk:
                sample_ids = np.take_along_axis(indices_k, sample_ids, axis=1)
            # 拼接到output
            output_ids = np.concatenate([output_ids, sample_ids], axis=1)
            end_counts = (output_ids == self.end_id).sum(1)  # 统计结束标记
            if output_ids.shape[1] >= self.minlen:  # 判断是否达到最短要求
                flag = end_counts >= min_ends  # 统计满足结束标记结果
                # 收集已结束句子并更新 inputs 和 output
                if flag.any():
                    for i in output_ids[flag]:
                        result.append(i)

                    remain_flag = flag == False
                    inputs = [i[remain_flag] for i in inputs]
                    output_ids = output_ids[remain_flag]
                # 没有剩余句子则跳出
                if len(output_ids) == 0:
                    break
        # 达到最大长度任然没有结束的直接添加进结果列表
        result.extend(output_ids)
        return result


def insert_arguments(**arguments):
    """类的方法上插入一个带有默认值的参数"""

    def decorator(func):
        def new_func(self, *args, **kwargs):
            for k, v in arguments.items():
                if k in kwargs:
                    v = kwargs.pop(k)  # 用户自定义则覆盖默认值
                setattr(self, k, v)
            return func(self, *args, **kwargs)

        return new_func

    return decorator


def remove_arguments(*argments):
    """类方法上禁用某些参数"""

    def decorator(func):
        def new_func(self, *args, **kwargs):
            for k in argments:
                if k in kwargs:
                    raise TypeError(
                        '%s got an unexpected keyword argument \'%s\'' %
                        (self.__class__.__name__, k))

            return func(self, *args, **kwargs)

        return new_func

    return decorator


class DataGenerator(object):
    """
    数据生成器，用于生成 批量 样本
    example:
    class CIFAR10Generator(DataGenerator):
            def __iter__(self):
                batch_x, batch_y = [], []
                for is_end, item in self.get_sample():
                    file_name, y = item
                    batch_x.append(resize(imread(file_name),(200,200))
                    batch_y.append(y)
                    if is_end or len(batch_x) == self.batch_size:
                        yield batch_x, batch_y
                        batch_x, batch_y = [], []
    cifar10_generate = (file_names_with_label, batch_size=32, shuffle=True)
    """

    def __init__(self, data, batch_size=32, buffer_size=None):
        """
        样本迭代器
        """
        self.data = data
        self.batch_size = batch_size
        if hasattr(data, '__len__'):
            self.steps = int(np.ceil(len(data) / float(batch_size)))
        else:
            self.steps = None
        self.buffer_size = buffer_size or batch_size * 1000

    def __len__(self):
        return self.steps

    def get_sample(self, shuffle=False):
        """
        gets one sample data with a flag of is this data is the last one
        """
        if shuffle:
            if self.steps is None:
                def generator():
                    cache, buffer_full = [], False
                    for item in self.data:
                        cache.append(item)
                        if buffer_full:
                            idx = np.random.randint(len(cache))
                            yield cache.pop(idx)
                        elif len(cache) == self.buffer_size:
                            buffer_full = True

                    while cache:
                        idx = np.random.randint(len(cache))
                        yield cache.pop(idx)
            else:
                def generator():
                    indices = list(range(len(self.data)))
                    np.random.shuffle(indices)
                    for idx in indices:
                        yield self.data[idx]

            data = generator()
        else:
            data = iter(self.data)

        current_data = next(data)
        for next_data in data:
            yield False, current_data
            current_data = next_data

        yield True, current_data

    @abstractmethod
    def __iter__(self, shuffle=False):
        """ 处理单个样本并构造batch data
        """
        raise NotImplementedError

    def generator(self):
        while True:
            for d in self.__iter__(shuffle=True):
                yield d

    def take(self, nums=1, shuffle=False):
        """take nums * batch examples"""
        d = []
        for i, data in enumerate(self.__iter__(shuffle)):
            if i >= nums:
                break

            d.append(data)

        if nums == 1:
            return d[0]
        return d


def pad_sequences(sequences, maxlen=None, value=0):
    """
    pad sequences (num_samples, num_timesteps) to same length
    """
    if maxlen is None:
        maxlen = max(len(x) for x in sequences)

    outputs = []
    for x in sequences:
        x = x[:maxlen]
        pad_range = (0, maxlen - len(x))
        x = np.pad(array=x, pad_width=pad_range, mode='constant', constant_values=value)
        outputs.append(x)
    return np.array(outputs)


class ViterbiDecoder(object):
    """viterbi 解码基类"""

    def __init__(self, trans, starts=None, ends=None):
        """
        :param trans:  转移矩阵
        :param starts:  开始标签index集合
        :param ends: 结束标签index集合
        :return:
        """
        self.trans = trans
        self.num_labels = len(trans)
        self.starts = starts
        self.ends = ends

        self.non_starts = []
        self.non_ends = []
        if starts is not None:
            all_labels = list(range(self.num_labels))
            self.non_starts = [label for label in all_labels if label not in starts]
        if ends is not None:
            all_labels = list(range(self.num_labels))
            self.non_ends = [label for label in all_labels if label not in starts]

    def decode(self, points):
        """points shape: (sequence_length, num_labels)"""
        points[0, self.non_starts] -= np.inf
        points[-1, self.non_ends] -= np.inf
        paths = np.arange(self.num_labels).reshape((-1, 1))
        score = points[0].reshape((-1, 1))
        labels = paths
        for idx in range(1, len(points)):
            all_scores = score + self.trans + points[idx].reshape((1, -1))
            max_idx = all_scores.argmax(0)
            score = all_scores.max(0).reshape((-1, 1))
            paths = np.concatenate([paths[max_idx, :], labels], axis=-1)

        return paths[score[:, 0].argmax(), :]


def text_segmentate(text, maxlen, seps='\n', strips=None):
    """过滤strips，按照seps顺序切分句子为若干个短句子"""
    text = text.strip().strip(strips)
    if seps and len(text) > maxlen:
        pieces = text.split(seps[0])
        text, texts = '', []
        for i, p in enumerate(pieces):
            if text and p and len(text) + len(p) > maxlen - 1:
                texts.extend(text_segmentate(text, maxlen, seps[1:], strips))
                text = ''
            if i + 1 == len(pieces):
                text = text + p
            else:
                text = text + p + seps[0]
        if text:
            texts.extend(text_segmentate(text, maxlen, seps[1:], strips))
        return texts
    else:
        return [text]


def string_matching(s, keywords):
    """
    判断是否有keyword 出现在s内
    """
    for k in keywords:
        if re.search(k, s):
            return True
    return False



# -*- coding: utf-8 -*-
# @Date    : 2020/7/6
# @Author  : mingming.xu
# @Email   : xv44586@gmail.com
# @File    : optimizers.py
import re

import numpy as np
import tensorflow as tf
from keras.optimizers import *

from toolkit4nlp.backend import keras, K, is_tf_keras, piecewise_linear
from toolkit4nlp.utils import *


class Adam(keras.optimizers.Optimizer):
    '''
    w_t = w_t-1 - update_t
    update_t = lr * m_t / sqrt(v_t)
    m_t = beta_1 * m_t-1 + (1 - beta_1) * g_t
    v_t = beta_2 * v_t-1 + (1 - beta_2) * g_t**2
    由于更新前期梯度较小，容易朝着0方向走，所以通常加一个bias correct来校正方向
    m_t_hat = m_t / (1 + beta_1**t)
    v_t_hat = v_t / (1 + beta_2 ** t)

    ref:
    - [zhihu-zhuanlan](
            https://zhuanlan.zhihu.com/p/32230623)
    - [Adam - A Method for Stochastic Optimization](
           https://arxiv.org/abs/1412.6980v8)
    - [On the Convergence of Adam and Beyond](
           https://openreview.net/forum?id=ryQu7f-RZ)

    '''

    def __init__(self, learning_rate=0.001, beta_1=0.9, beta_2=0.99, epsilon=1e-6, bias_correct=True, **kwargs):
        kwargs['name'] = kwargs.get('name', 'Adam')
        super(Adam, self).__init__(**kwargs)
        self._set_hyper('learning_rate', learning_rate)
        self._set_hyper('beta_1', beta_1)
        self._set_hyper('beta_2', beta_2)
        self.epsilon = epsilon or K.epsilon()
        self.bias_correct = bias_correct

    def _create_slots(self, var_list):
        for var in var_list:
            self.add_slot(var, 'm')
            self.add_slot(var, 'v')

    def _resource_apply(self, grad, var, indices=None):
        var_dtype = var.dtype.base_dtype
        lr_t = self._decayed_lr(var_dtype)
        m = self.get_slot(var, 'm')
        v = self.get_slot(var, 'v')
        beta_1_t = self._get_hyper('beta_1', var_dtype)
        beta_2_t = self._get_hyper('beta_2', var_dtype)
        local_step = K.cast(self.iterations + 1, var_dtype)
        beta_1_power = K.pow(beta_1_t, local_step)
        beta_2_power = K.pow(beta_2_t, local_step)

        # update
        if indices is None:
            m_t = K.update(m, beta_1_t * m + (1 - beta_1_t) * grad)
            v_t = K.update(v, beta_2_t * v + (1 - beta_2_t) * grad ** 2)
        else:
            mv_ops = [K.update(m, beta_1_t * m), K.update(v, beta_2_t * v)]
            with tf.control_dependencies(mv_ops):
                m_t = self._resource_scatter_add(m, indices, (1 - beta_1_t) * grad)
                v_t = self._resource_scatter_add(v, indices, (1 - beta_2_t) * grad ** 2)

        #
        with tf.control_dependencies([m_t, v_t]):
            if self.bias_correct:
                m_t = m_t / (1 + beta_1_power)
                v_t = v_t / (1 + beta_2_power)
            var_t = var - lr_t * m_t / (K.sqrt(v_t) + self.epsilon)
            return K.update(var, var_t)

    def _resource_apply_dense(self, grad, var):
        return self._resource_apply(grad, var)

    def _resource_apply_sparse(self, grad, var, indices):
        return self._resource_apply(grad, var, indices)

    def get_config(self):
        config = {
            'learnint_rate': self._serialize_hyperparameter('learning_rate'),
            'beta_1': self._serialize_hyperparameter('beta_1'),
            'beta_2': self._serialize_hyperparameter('beta_2'),
            'epsilon': self.epsilon
        }
        basic_config = super(Adam, self).get_config()
        return dict(list(basic_config.items()) + list(config.items()))


class AdaBelief(keras.optimizers.Optimizer):
    """AdaBelief optimizer.

    Default parameters follow those provided in the original paper.

    # Arguments
        learning_rate: float >= 0. Learning rate.
        beta_1: float, 0 < beta < 1. Generally close to 1.
        beta_2: float, 0 < beta < 1. Generally close to 1.
        amsgrad: boolean. Whether to apply the AMSGrad variant of this
            algorithm from the paper "On the Convergence of Adam and
            Beyond".

    # References
        - [Adam - A Method for Stochastic Optimization](
           https://arxiv.org/abs/1412.6980v8)
        - [AdaBelief Optimizer: Adapting Stepsizes by the Belief in Observed Gradients](
           https://arxiv.org/pdf/2010.07468.pdf)
    """

    def __init__(self, learning_rate=0.001, beta_1=0.9, beta_2=0.999,
                 amsgrad=False, **kwargs):
        self.initial_decay = kwargs.pop('decay', 0.0)
        self.epsilon = kwargs.pop('epsilon', K.epsilon())
        learning_rate = kwargs.pop('lr', learning_rate)
        super(AdaBelief, self).__init__(**kwargs)
        with K.name_scope(self.__class__.__name__):
            self.iterations = K.variable(0, dtype='int64', name='iterations')
            self.learning_rate = K.variable(learning_rate, name='learning_rate')
            self.beta_1 = K.variable(beta_1, name='beta_1')
            self.beta_2 = K.variable(beta_2, name='beta_2')
            self.decay = K.variable(self.initial_decay, name='decay')
        self.amsgrad = amsgrad

    @K.symbolic
    def get_updates(self, loss, params):
        grads = self.get_gradients(loss, params)
        self.updates = [K.update_add(self.iterations, 1)]

        lr = self.learning_rate
        if self.initial_decay > 0:
            lr = lr * (1. / (1. + self.decay * K.cast(self.iterations,
                                                      K.dtype(self.decay))))

        t = K.cast(self.iterations, K.floatx()) + 1
        lr_t = lr * (K.sqrt(1. - K.pow(self.beta_2, t)) /
                     (1. - K.pow(self.beta_1, t)))

        ms = [K.zeros(K.int_shape(p),
              dtype=K.dtype(p),
              name='m_' + str(i))
              for (i, p) in enumerate(params)]
        vs = [K.zeros(K.int_shape(p),
              dtype=K.dtype(p),
              name='v_' + str(i))
              for (i, p) in enumerate(params)]

        if self.amsgrad:
            vhats = [K.zeros(K.int_shape(p),
                     dtype=K.dtype(p),
                     name='vhat_' + str(i))
                     for (i, p) in enumerate(params)]
        else:
            vhats = [K.zeros(1, name='vhat_' + str(i))
                     for i in range(len(params))]
        self.weights = [self.iterations] + ms + vs + vhats

        for p, g, m, v, vhat in zip(params, grads, ms, vs, vhats):
            m_t = (self.beta_1 * m) + (1. - self.beta_1) * g
            v_t = (self.beta_2 * v) + (1. - self.beta_2) * K.square(g - m_t)
            if self.amsgrad:
                vhat_t = K.maximum(vhat, v_t)
                p_t = p - lr_t * m_t / (K.sqrt(vhat_t) + self.epsilon)
                self.updates.append(K.update(vhat, vhat_t))
            else:
                p_t = p - lr_t * m_t / (K.sqrt(v_t) + self.epsilon)

            self.updates.append(K.update(m, m_t))
            self.updates.append(K.update(v, v_t))
            new_p = p_t

            # Apply constraints.
            if getattr(p, 'constraint', None) is not None:
                new_p = p.constraint(new_p)

            self.updates.append(K.update(p, new_p))
        return self.updates

    def get_config(self):
        config = {'learning_rate': float(K.get_value(self.learning_rate)),
                  'beta_1': float(K.get_value(self.beta_1)),
                  'beta_2': float(K.get_value(self.beta_2)),
                  'decay': float(K.get_value(self.decay)),
                  'epsilon': self.epsilon,
                  'amsgrad': self.amsgrad}
        base_config = super(AdaBelief, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class AdaBeliefTf(keras.optimizers.Optimizer):
    """tf.keras 版
    """

    def __init__(
            self,
            learning_rate=0.001,
            beta_1=0.9,
            beta_2=0.999,
            epsilon=1e-6,
            bias_correct=True,
            **kwargs
    ):
        kwargs['name'] = kwargs.get('name') or 'AdaBelief'
        super(AdaBeliefTf, self).__init__(**kwargs)
        self._set_hyper('learning_rate', learning_rate)
        self._set_hyper('beta_1', beta_1)
        self._set_hyper('beta_2', beta_2)
        self.epsilon = epsilon or K.epislon()
        self.bias_correct = bias_correct

    def _create_slots(self, var_list):
        for var in var_list:
            self.add_slot(var, 'm')
            self.add_slot(var, 'v')

    def _resource_apply(self, grad, var, indices=None):
        # 准备变量
        var_dtype = var.dtype.base_dtype
        lr_t = self._decayed_lr(var_dtype)
        m = self.get_slot(var, 'm')
        v = self.get_slot(var, 'v')
        beta_1_t = self._get_hyper('beta_1', var_dtype)
        beta_2_t = self._get_hyper('beta_2', var_dtype)
        epsilon_t = K.cast(self.epsilon, var_dtype)
        local_step = K.cast(self.iterations + 1, var_dtype)
        beta_1_t_power = K.pow(beta_1_t, local_step)
        beta_2_t_power = K.pow(beta_2_t, local_step)

        # 更新公式

        m_t = K.update(m, beta_1_t * m + (1 - beta_1_t) * grad)
        v_t = K.update(v, beta_2_t * v + (1 - beta_2_t) * (grad - m_t) ** 2)

        # 返回算子
        with tf.control_dependencies([m_t, v_t]):
            if self.bias_correct:
                m_t = m_t / (1.0 - beta_1_t_power)
                v_t = v_t / (1.0 - beta_2_t_power)
            var_t = var - lr_t * m_t / (K.sqrt(v_t) + epsilon_t)
            return K.update(var, var_t)

    def _resource_apply_dense(self, grad, var):
        return self._resource_apply(grad, var)

    def _resource_apply_sparse(self, grad, var, indices):
        grad = tf.IndexedSlices(grad, indices, K.shape(var))
        grad = tf.convert_to_tensor(grad)
        return self._resource_apply_dense(grad, var)

    def get_config(self):
        config = {
            'learning_rate': self._serialize_hyperparameter('learning_rate'),
            'beta_1': self._serialize_hyperparameter('beta_1'),
            'beta_2': self._serialize_hyperparameter('beta_2'),
            'epsilon': self.epsilon,
            'bias_correct': self.bias_correct,
        }
        base_config = super(AdaBeliefTf, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


def export_to_custom_objects(extend_with_func):
    def new_extend_with_func(BaseOptimizer, name=None):
        NewOptimizer = extend_with_func(BaseOptimizer)
        if name:
            NewOptimizer.__name__ = name
        name = NewOptimizer.__name__
        keras.utils.get_custom_objects()[name] = NewOptimizer
        return NewOptimizer

    return new_extend_with_func


@export_to_custom_objects
def extend_with_gradient_accumulation_tf2(BaseOptimizer):
    class NewOptimizer(BaseOptimizer):
        @insert_arguments(grad_accum_steps=2)
        def __init__(self, *args, **kwargs):
            super(NewOptimizer, self).__init__(*args, **kwargs)

        def _create_slots(self, var_list):
            super(NewOptimizer, self)._create_slots(var_list)
            for var in var_list:
                self.add_slot(var, 'gradient_accumulation')

        def _resource_apply(self, grad, var, indices=None):
            """interation % acc_steps==0 then update else accumulate
               思路是先判断是否累计了 acc_steps，如果没有，则update时保持原样，
               并累计当前梯度，否则，更新梯度并将累计的梯度置零
            """
            #  是否更新
            cond = K.equal(self.iterations % self.grad_accum_steps, 0)
            #  获取梯度累计量
            gradient_accumulation = self.get_slot(var, 'gradient_accumulation')

            # 获取平均梯度
            gradient_t = gradient_accumulation / self.grad_accum_steps

            old_update = K.update

            # 根据条件判断是否真的更新
            def new_update(x, new_x):
                new_x = K.switch(cond, new_x, x)
                return old_update(x, new_x)

            K.update = new_update
            op = super(NewOptimizer, self)._resource_apply(gradient_t, var)
            K.update = old_update

            # 根据条件判断是否需要置零
            with tf.control_dependencies([op]):
                gradient_t = K.switch(cond, K.zeros_like(gradient_accumulation), gradient_accumulation)
                with tf.control_dependencies([K.update(gradient_accumulation, gradient_t)]):
                    if indices is None:
                        gradient_t = K.update(gradient_accumulation, gradient_accumulation + grad)
                    else:
                        gradient_t = self._resource_scatter_add(gradient_accumulation, indices, grad)
            return gradient_t

        def get_config(self):
            config = super(NewOptimizer, self).get_config()
            config.update({'grad_accum_steps': self.grad_accum_steps})
            return config

    return NewOptimizer


@export_to_custom_objects
def extend_with_gradient_accumulation(BaseOptimizer):
    """原生keras版"""

    class NewOptimizer(BaseOptimizer):
        @insert_arguments(grad_accum_steps=2)
        def __init__(self, *args, **kwargs):
            super(NewOptimizer, self).__init__(*args, **kwargs)
            self._first_grad = True  # first grad

        @K.symbolic
        def get_updates(self, loss, params):
            # 是否更新
            cond = K.equal(self.iterations % self.grad_accum_steps, 0)
            cond = K.cast(cond, K.floatx())
            # 获取梯度
            grads = self.get_gradients(loss, params)
            self.accum_grads = [K.zeros(
                shape=K.int_shape(p), dtype=K.dtype(p), name='accum_grad_{}'.format(i)) for i, p in enumerate(params)]

            old_update = K.update

            def new_update(x, new_x):
                new_x = cond * new_x + (1 - cond) * x
                return old_update(x, new_x)

            K.update = new_update
            updates = super(NewOptimizer, self).get_updates(loss, params)
            K.update = old_update

            # 累计更新
            with K.control_dependencies(updates):
                acc_updates = [
                    K.update(ag, g + (1 - cond) * ag) for ag, g in zip(self.accum_grads, grads)
                ]

            return acc_updates

        def get_gradients(self, loss, params):
            if self._first_grad:
                self._first_grad = False
                return super(NewOptimizer, self).get_gradients(loss, params)
            else:
                return [ag / self.grad_accum_steps for ag in self.accum_grads]

        def get_config(self):
            config = {'grad_accum_steps': self.grad_accum_steps}
            base_config = super(NewOptimizer, self).get_config()
            return dict(list(base_config.items()) + list(config.items()))

    return NewOptimizer


@export_to_custom_objects
def extend_with_weight_decay_tf2(BaseOptimizer):
    """增加权重衰减
    ref: [DECOUPLED WEIGHT DECAY REGULARIZATION](https://arxiv.org/pdf/1711.05101.pdf)
    大多数框架在实现L2 regularization时是使用weight decay，然而L2 regularization 与 weight decay 在标准 SGD下是等价的，
    但是当使用Adam时，缺不是等价的，原因是：
    g_t = ▽f_t-1 + λθ，其中λθ是 L2 loss的梯度
    m_t = β_1 * m_t-1 + (1 - β_1) * g_t
    v_t = β_2 * v_t-2 + (1 - β_2) * g_t^2
    θ_t = θ_t - 1 - α(m_t / v_t^0.5 + ε)
    代入上面三式后带有θ的项为 α（λθ/ v_t^0.5 + ε),这导致梯度变化越大的方向，权重约束越小，这显然不合理。
    L2 regularization应该是各向同性。一种改进这个问题的方法就是将梯度下降与weight decay 解耦，
    不在求梯度时代入weight decay ，而是在整个梯度下降完成后，加入weight decay，这样将梯度下降与weight decay解耦，
    达到L2 regularization效果
    """

    class NewOptimizer(BaseOptimizer):
        @insert_arguments(weight_decay_rate=0.01, exclude_from_weight_decay=[])
        def __init__(self, *args, **kwargs):
            super(NewOptimizer, self).__init__(*args, **kwargs)

        def _resource_apply(self, grad, var, indices=None):
            old_update = K.update

            def new_update(x, new_x):
                if x is var and self._do_use_weight_decay(x):
                    lr_t = self._decayed_lr(x.dtype.base_dtype)
                    new_x = new_x - lr_t * self.weight_decay_rate * x
                return old_update(x, new_x)

            K.update = new_update
            op = super(NewOptimizer, self)._resource_apply(grad, var, indices)
            K.update = old_update
            return op

        def _do_use_weight_decay(self, param):
            """Whether to use L2 weight decay for `param_name`."""
            param_name = param.name
            if not self.weight_decay_rate:
                return False
            if self.exclude_from_weight_decay:
                for r in self.exclude_from_weight_decay:
                    if re.search(r, param_name) is not None:
                        return False
            return True

        def get_config(self):
            config = super(NewOptimizer, self).get_config()
            config.update({'weight_decay_rate': self.weight_decay_rate,
                           'exclude_from_weight_decay': self.exclude_from_weight_decay})
            return config

    return NewOptimizer


@export_to_custom_objects
def extend_with_weight_decay(BaseOptimizer):
    """原生keras版"""

    class NewOptimizer(BaseOptimizer):
        @insert_arguments(weight_decay_rate=0.01, exclude_from_weight_decay=[])
        def __init__(self, *args, **kwargs):
            super(NewOptimizer, self).__init__(*args, **kwargs)
            if not hasattr(self, 'learning_rate'):
                self.learning_rate = self.lr

        @K.symbolic
        def get_update(self, loss, params):
            old_update = K.update

            def new_update(x, new_x):
                if x in params and self._do_use_weight_decay(x):
                    new_x = new_x - self.learning_rate * self.weight_decay_rate * x
                return old_update(x, new_x)

            K.update = new_update
            updates = super(NewOptimizer, self).get_update(loss, params)
            K.update = old_update
            return updates

        def _do_use_weight_decay(self, param):
            """Whether to use L2 weight decay for `param_name`."""
            param_name = param.name
            if not self.weight_decay_rate:
                return False
            if self.exclude_from_weight_decay:
                for r in self.exclude_from_weight_decay:
                    if re.search(r, param_name) is not None:
                        return False
            return True

        def get_config(self):
            config = {'weight_decay_rate': self.weight_decay_rate,
                      'exclude_from_weight_decay': self.exclude_from_weight_decay}
            base_config = super(NewOptimizer, self).get_config()
            return dict(list(base_config.items()) + list(config.items()))

    return NewOptimizer


@export_to_custom_objects
def extend_with_piecewise_linear_lr_tf2(BaseOptimizer):
    """
    分段线性学习率，使用场景如 warmup
    """

    class NewOptimzer(BaseOptimizer):
        """
        schedule 是一个{ point: value} 的字典，如 {10: 1, 20: 0.5}代表从 0 到 10 步 lr 从 0 线性增加到 100% ，
        然后从 10 到 20 线性降低到 50%，之后一直保持 50% 不变
        """

        @insert_arguments(lr_schedule={0: 1})
        def __init__(self, *args, **kwargs):
            super(NewOptimzer, self).__init__(*args, **kwargs)
            self.lr_schedule = {int(t): v for t, v in self.lr_schedule.items()}

        def _decayed_lr(self, var_dtypes):
            """重写获取decayed learning rate 方法"""

            lr_t = super(NewOptimzer, self)._decayed_lr(var_dtypes)
            lr_rate = piecewise_linear(self.iterations, self.lr_schedule)
            return lr_t * K.cast(lr_rate, var_dtypes)

        def get_config(self):
            config = super(NewOptimzer, self).get_config()
            config.update({'lr_schedule': self.lr_schedule})
            return config

    return NewOptimzer


@export_to_custom_objects
def extend_with_piecewise_linear_lr(BaseOptimizer):
    """原生keras版"""
    class NewOptimizer(BaseOptimizer):
        @insert_arguments(lr_schedule={0:1})
        def __init__(self, *args, **kwargs):
            super(NewOptimizer, self).__init__(*args, **kwargs)
            self.lr_schedule = {int(t): v for t, v in self.lr_schedule.items()}

        @K.symbolic
        def get_update(self, loss, params):
            # 获取当前 step 的 lr rate
            lr_rate_t = piecewise_linear(self.iterations, self.lr_schedule)

            old_update = K.update

            def new_update(x, new_x):
                new_x = x + (new_x - x) * lr_rate_t  # 按照当前lr rate 缩放 update
                return old_update(x, new_x)

            K.update = new_update
            updates = super(NewOptimizer, self).get_update(loss, params)
            K.update = old_update
            return updates

        def get_config(self):
            config = {'lr_schedule': self.lr_schedule}
            base_config = super(NewOptimizer, self).get_config()
            return dict(list(base_config.items()) + list(config.items()))

    return NewOptimizer


# keras or tf.keras
if is_tf_keras:
    extend_with_piecewise_linear_lr = extend_with_piecewise_linear_lr_tf2
    extend_with_gradient_accumulation = extend_with_gradient_accumulation_tf2
    extend_with_weight_decay = extend_with_weight_decay_tf2
    AdaBelief = AdaBeliefTf
else:
    Adam = keras.optimizers.Adam

custom_objects = {
    'Adam': Adam,
    'AdaBelief': AdaBelief,
}

keras.utils.get_custom_objects().update(custom_objects)



# -*- coding: utf-8 -*-
# @Date    : 2020/7/6
# @Author  : mingming.xu
# @Email   : xv44586@gmail.com
# @File    : backend.py
# 判断是tf.keras还是纯keras的标记
import os, sys
from distutils.util import strtobool
import numpy as np
import tensorflow as tf

is_tf_keras = strtobool(os.environ.get('TF_KERAS', '0'))

if is_tf_keras:
    import tensorflow.keras as keras
    import tensorflow.keras.backend as K

    sys.modules['keras'] = keras
else:
    import keras
    import keras.backend as K


def gelu_tanh(x):
    """Gaussian Error Linear Unit.

  This is a smoother version of the RELU.
  Original paper: https://arxiv.org/abs/1606.08415
  Args:
    x: float Tensor to perform activation.

  Returns:
    `x` with the GELU activation applied.
  """
    cdf = 0.5 * (1.0 + K.tanh(
        (np.sqrt(2 / np.pi) * (x + 0.044715 * K.pow(x, 3)))))
    return x * cdf


def gelu_erf(x):
    """基于Erf直接计算的gelu函数
    """
    return 0.5 * x * (1.0 + tf.math.erf(x / np.sqrt(2.0)))


def set_gelu_version(version_str):
    if version_str not in ['gelu_erf', 'gelu_tanh']:
        raise ValueError('{} version not supported.'.format(version_str))
    if version_str == 'gelu_erf':
        keras.utils.get_custom_objects().update({'gelu': gelu_erf()})

    keras.utils.get_custom_objects().update({'gelu': gelu_tanh()})


def sequence_masking(x, mask, mode=0, axis=1):
    '''
    mask shape: [batch_size, seq_length]
    :param x:
    :param mask: 0,1 矩阵
    :param mode: 直接相乘 or 减大数模拟
    :param axis:
    :return:
    '''
    if mask is None or mode not in [0, 1, 'mul', 'add']:
        return x

    if axis == -1:
        axis = K.ndim(x) - 1
    assert axis > 0, 'axis must greater than 0'
    # 形状扩充
    # left
    for _ in range(axis - 1):
        mask = K.expand_dims(mask, 1)
    # right
    for _ in range(K.ndim(x) - K.ndim(mask) - axis + 1):
        mask = K.expand_dims(mask, K.ndim(mask))

    if mode in [0, 'mul']:
        return x * mask

    return x - (1 - mask) * 1e12


def piecewise_linear(global_steps, lr_schedule):
    schedule = sorted(lr_schedule.items())
    if schedule[0][0] != 0:
        schedule = [(0, 0.0)] + schedule  # 增加起点

    v = K.cast(schedule[0][1], K.floatx())
    p = K.cast(global_steps, K.floatx())

    for i in range(len(schedule)):
        p_begin = schedule[i][0]
        v_begin = v
        if i != len(schedule) - 1:
            point_range = schedule[i + 1][0] - schedule[i][0]
            value_range = schedule[i + 1][1] - schedule[i][1]
            rate = 1.0 * value_range / point_range
            v = schedule[i][1] + rate * (p - p_begin)
        else:
            v = K.cast(schedule[i][1], K.floatx())

        v = K.switch(p > p_begin, v, v_begin)
    return v

def swish(x):
    """swish函数（这样封装过后才有 __name__ 属性）
    """
    return tf.nn.swish(x)


def symbolic(f):
    """恒等装饰器（兼容旧版本keras用）
    """
    return f


def leaky_relu(x, alpha=0.2):
    """leaky relu函数（这样封装过后才有 __name__ 属性）
    """
    return tf.nn.leaky_relu(x, alpha=alpha)


def search_layer(inputs, name, exclude_from=None):
    """根据inputs和name来搜索层
    说明：inputs为某个层或某个层的输出；name为目标层的名字。
    实现：根据inputs一直往上递归搜索，直到发现名字为name的层为止；
         如果找不到，那就返回None。
    """
    if exclude_from is None:
        exclude_from = set()

    if isinstance(inputs, keras.layers.Layer):
        layer = inputs
    else:
        layer = inputs._keras_history[0]

    if layer.name == name:
        return layer
    elif layer in exclude_from:
        return None
    else:
        exclude_from.add(layer)
        if isinstance(layer, keras.models.Model):
            model = layer
            for layer in model.layers:
                if layer.name == name:
                    return layer
        inbound_layers = layer._inbound_nodes[0].inbound_layers
        if not isinstance(inbound_layers, list):
            inbound_layers = [inbound_layers]
        if len(inbound_layers) > 0:
            for layer in inbound_layers:
                layer = search_layer(layer, name, exclude_from)
                if layer is not None:
                    return layer


# focal loss for binary classification
def binary_focal_loss(alpha=0.25, gamma=2.):
    """
    (focal loss for binary classification)[https://arxiv.org/pdf/1708.02002.pdf]
    FL(p) = \left\{\begin{matrix}
         -\alpha * (1-p)^{\gamma } *log(p),& if &  y=1\\
         -(1-\alpha * (p)^{\gamma} * log(1-p),& if & y=0
        \end{matrix}\right.
    """
    def focal_loss(y_true, y_pred):
        y_true = K.cast(y_true, 'int32')
        p1 = K.switch(K.equal(y_true, 1), y_pred, K.ones_like(y_pred))
        p0 = K.switch(K.equal(y_true, 0), y_pred, K.zeros_like(y_pred))
        return - K.mean(alpha * K.pow(1 - p1, gamma) * K.log(p1)) - K.mean(
            (1 - alpha) * K.pow(p0, gamma) * K.log(1 - p0))

    return focal_loss


# 给旧版本keras新增symbolic方法（装饰器），
# 以便兼容optimizers.py中的代码
K.symbolic = getattr(K, 'symbolic', None) or symbolic
custom_objects = {
    'gelu_erf': gelu_erf,
    'gelu_tanh': gelu_tanh,
    'gelu': gelu_erf,
    'swish': swish,
    'leaky_relu': leaky_relu,
    'binary_focal_loss': binary_focal_loss,
}

keras.utils.get_custom_objects().update(custom_objects)



#!/usr/bin/env python
# -*- coding: utf8 -*-
# @Date     : 2020/06/01
# @Author   : mingming.xu
# @Email    : xv44586@gmail.com
import numpy as np
import tensorflow as tf

from toolkit4nlp.backend import K, keras
from toolkit4nlp.backend import sequence_masking
from toolkit4nlp import initializers
from keras.layers import *
from keras import activations


class Layer(keras.layers.Layer):
    # 所有自定义layer都支持masking
    def __init__(self, **kwargs):
        super(Layer, self).__init__(**kwargs)
        self.supports_masking = True


class MultiHeadAttention(Layer):
    """
    多头注意力机制
    """

    def __init__(self, head_nums, head_size, key_size=None, use_bias=True, attention_scale=True,
                 kernel_initializer='glorot_uniform', with_residual_attention=False, **kwargs):
        super(MultiHeadAttention, self).__init__(**kwargs)
        self.head_nums = head_nums
        self.head_size = head_size
        self.key_size = key_size or head_size
        self.output_dim = head_nums * head_size
        self.use_bias = use_bias
        self.attention_scale = attention_scale
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.with_residual_attention = with_residual_attention  # realformer

    def build(self, input_shape):
        super(MultiHeadAttention, self).build(input_shape)
        self.q_dense = Dense(
            units=self.key_size * self.head_nums,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer
        )
        self.k_dense = Dense(
            units=self.key_size * self.head_nums,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer
        )
        self.v_dense = Dense(
            units=self.output_dim,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer
        )

        self.combine_dense = Dense(
            units=self.output_dim,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer
        )

    def call(self, inputs, mask=None, **kwargs):
        """
        多头注意力
        :param inputs: [q, k, v, a_bias, position_bias]
        :param mask: [q_mask, k_mask, v_mask],
            q_mask 对query序列进行mask，针对padding；v_mask对value序列进行mask，防止看到某些位置value，如padding
        :param a_bias: Boolean，是否对attention进行偏移，包括attention mask/ pre attention scores
        :param position_bias: type of position bias， 使用指定类型的位置编码对attention里的位置信息进行偏移
        :return:
        """
        q, k, v = inputs[:3]
        q_mask, v_mask, idx = None, None, 3
        if mask is not None:
            if mask[0] is not None:
                q_mask = K.cast(mask[0], K.floatx())
            if mask[2] is not None:
                v_mask = K.cast(mask[2], K.floatx())
        attention_bias, position_bias = kwargs.get('a_bias'), kwargs.get('position_bias')
        if attention_bias:
            attention_bias = inputs[idx]
            idx += 1

        # 投影变换
        qw = self.q_dense(q)
        kw = self.k_dense(k)
        vw = self.v_dense(v)

        # 形状变换
        qw = K.reshape(qw, [-1, K.shape(q)[1], self.head_nums, self.key_size])
        kw = K.reshape(kw, [-1, K.shape(k)[1], self.head_nums, self.key_size])
        vw = K.reshape(vw, [-1, K.shape(v)[1], self.head_nums, self.head_size])
        # 计算attention
        att = tf.einsum('bjhd,bkhd->bhjk', qw, kw)
        # 处理位置编码
        if position_bias == 'relative':
            position_embeddings = inputs[idx]
            att = att + tf.einsum('bjhd,jkd->bhjk', qw, position_embeddings)

        if self.attention_scale:
            att = att / self.key_size ** 0.5

        # value mask
        att = sequence_masking(att, v_mask, 'add', -1)

        # attention mask / residual attention scores
        if attention_bias is not None:
            att += attention_bias
        attention_bias = att

        att = K.softmax(att)
        output = tf.einsum('bhjk,bkhd->bjhd', att, vw)
        # 继续处理位置编码
        if position_bias == 'relative':
            output = output + tf.einsum('bhjk,jkd->bjhd', att, position_embeddings)
        output = K.reshape(output, (-1, K.shape(output)[1], self.output_dim))
        output = self.combine_dense(output)

        # query mask
        # output = sequence_masking(output, q_mask, 'mul')
        if not self.with_residual_attention:
            return output

        return [output, attention_bias]

    def compute_output_shape(self, input_shape):
        output_shape = (input_shape[0][0], input_shape[0][1], self.output_dim)
        if self.with_residual_attention:
            att_shape = (input_shape[0][0], self.head_nums, input_shape[0][1], input_shape[1][1])
            return [output_shape, att_shape]

        return output_shape

    def compute_mask(self, inputs, mask=None):
        if mask is not None:
            if self.with_residual_attention:
                return [mask[0], None]
            return mask[0]

    def get_config(self):
        config = super(MultiHeadAttention, self).get_config()
        config.update({'head_nums': self.head_nums,
                       'head_size': self.head_size,
                       'key_size': self.key_size,
                       'use_bias': self.use_bias,
                       'attention_scale': self.attention_scale,
                       'with_residual_attention': self.with_residual_attention,
                       'kernel_initializer': initializers.serialize(self.kernel_initializer)})
        return config


class LayerNormalization(Layer):
    """
    [Layer Normalization](https://arxiv.org/pdf/1607.06450.pdf)
    x_mean = mean(x)
    std = sqrt(mean(square(x - x_mean)) + epsilon)
    x = beta + (x - x_mean / std) * gamma
    """
    """
    增加Conditional，用来控制模型的表现。主要思路为通过调整center 与scale，来调整模型的输出，控制其表现。
    condition_* 为Conditional 系列参数，当Conditional=True时生效。
    ref: [基于Conditional Layer Normalization的条件文本生成](https://kexue.fm/archives/7124)
    """

    def __init__(self, center=True, scale=True, epsilon=None,
                 conditional=False, condition_hidden_units=None, condition_hidden_activation='linear',
                 condition_hidden_initializer='glorot_uniform', **kwargs):
        super(LayerNormalization, self).__init__(**kwargs)
        self.center = center
        self.scale = scale
        self.epsilon = epsilon or 1e-12
        self.conditional = conditional
        self.condition_hidden_units = condition_hidden_units
        self.condition_hidden_activation = activations.get(condition_hidden_activation)
        self.condition_hidden_initializer = initializers.get(condition_hidden_initializer)

    def build(self, input_shape):
        super(LayerNormalization, self).build(input_shape)

        if self.conditional:
            shape = (input_shape[0][-1], )
        else:
            shape = (input_shape[-1], )

        if self.center:
            self.beta = self.add_weight(shape=shape, initializer='zeros', name='beta')
        if self.scale:
            self.gamma = self.add_weight(shape=shape, initializer='ones', name='gamma')

        if self.conditional:
            # hidden units 不为None，做一次维度转换
            if self.condition_hidden_units is not None:
                self.condition_hidden_dense = Dense(units=self.condition_hidden_units,
                                                    activation=self.condition_hidden_activation,
                                                    use_bias=False,
                                                    kernel_initializer=self.condition_hidden_initializer)
            # 用0初始化center_dense 与scale_dense ,防止扰乱原模型
            if self.center:
                self.beta_dense = Dense(units=shape[0], use_bias=False, kernel_initializer='zeros')
            if self.scale:
                self.gamma_dense = Dense(units=shape[0], use_bias=False, kernel_initializer='zeros')

    def call(self, inputs):
        """
        conditional 时， condition 放在inputs后面，[inputs, condition]
        """
        if self.conditional:
            inputs, cond = inputs
            if self.condition_hidden_units is not None:
                cond = self.condition_hidden_dense(cond)
            # 适配cond维度，与inputs保持一致
            for _ in range(K.ndim(inputs) - K.ndim(cond)):
                cond = K.expand_dims(cond, 1)
            if self.center:
                beta = self.beta_dense(cond) + self.beta
            if self.scale:
                gamma = self.gamma_dense(cond) + self.gamma
        else:
            beta = self.beta
            gamma = self.gamma

        output = inputs
        if self.center:
            mean = K.mean(inputs, axis=-1, keepdims=True)
            output = output - mean

        if self.scale:
            var = K.mean(K.square(output), axis=-1, keepdims=True)
            std = K.sqrt(var + self.epsilon)
            output = output / std
            output = output * gamma

        if self.center:
            output = output + beta

        return output

    def compute_mask(self, inputs, mask=None):
        if self.conditional:
            masks = mask if mask is not None else []
            masks = [K.expand_dims(m, 0) for m in masks if m is not None]
            if len(masks) == 0:
                return None
            else:
                return K.all(K.concatenate(masks, axis=0), axis=0)

        return mask

    def compute_output_shape(self, input_shape):
        if self.conditional:
            return input_shape[0]
        else:
            return input_shape

    def get_config(self):
        base_config = super(LayerNormalization, self).get_config()
        base_config.update({"center": self.center,
                            "scale": self.scale,
                            "epsilon": self.epsilon,
                            "conditional": self.conditional,
                            "condition_hidden_units": self.condition_hidden_units,
                            "condition_hidden_activation": activations.serialize(self.condition_hidden_activation),
                            "condition_hidden_initializer": initializers.serialize(self.condition_hidden_initializer)})
        return base_config


class TokenAndPositionEmbedding(Layer):
    def __init__(self, vocab_size, embed_dim, maxlen, **kwargs):
        super(TokenAndPositionEmbedding, self).__init__(**kwargs)
        self.vocab_size = vocab_size
        self.embed_dim = embed_dim
        self.maxlen = maxlen

    def build(self, input_shape):
        super(TokenAndPositionEmbedding, self).build(input_shape)
        self.token_emb = Embedding(input_dim=self.vocab_size, output_dim=self.embed_dim, name='token_emb')
        self.pos_emb = Embedding(input_dim=self.maxlen, output_dim=self.embed_dim, name='pos_emb')

    def call(self, inputs):
        maxlen = K.shape(inputs)[-1]
        token_emb = self.token_emb(inputs)
        pos = tf.range(start=0, limit=maxlen, delta=1)
        pos_emb = self.pos_emb(pos)
        return token_emb + pos_emb

    def compute_output_shape(self, input_shape):
        return tuple(list(input_shape) + [self.embed_dim])

    def get_config(self):
        base_config = super(TokenAndPositionEmbedding, self).get_config()
        config = {
            "vocab_size": self.vocab_size,
            "maxlen": self.maxlen,
            "embed_dim": self.embed_dim
        }
        return dict(list(base_config.items()) + list(config.items()))


class PositionEmbedding(Layer):
    """
    将位置Embedding与inputs 进行merge, merge mode : [add, concat].
    """

    def __init__(self,
                 input_dim,
                 output_dim,
                 merge_mode='add',
                 embeddings_initializer='zero',
                 **kwargs):
        super(PositionEmbedding, self).__init__(**kwargs)
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.merge_mode = merge_mode
        self.embeddings_initializer = initializers.get(embeddings_initializer)

    def build(self, input_shape):
        super(PositionEmbedding, self).build(input_shape)
        self.embeddings = self.add_weight(name='position_embedding',
                                          shape=(self.input_dim, self.output_dim),
                                          initializer=self.embeddings_initializer)

    def call(self, inputs):
        input_shape = K.shape(inputs)
        batch_size, seq_length = input_shape[0], input_shape[1]
        pos_embedding = self.embeddings[:seq_length]
        pos_embedding = K.expand_dims(pos_embedding, 0)
        if self.merge_mode != 'add':
            pos_embedding = K.tile(pos_embedding, [batch_size, 1, 1])

        if self.merge_mode == 'add':
            return inputs + pos_embedding
        return K.concatenate([inputs, pos_embedding], axis=-1)

    def compute_output_shape(self, input_shape):
        if self.merge_mode == 'add':
            return input_shape

        return input_shape[:2] + (input_shape[2] + self.output_dim,)

    def get_config(self):
        base_config = super(PositionEmbedding, self).get_config()
        config = {'input_dim': self.input_dim,
                  'output_dim': self.output_dim,
                  'merge_mode': self.merge_mode,
                  'embeddings_initializer': initializers.serialize(self.embeddings_initializer)}
        return dict(list(base_config.items()) + list(config.items()))


class FeedForward(Layer):
    """
    Dense(units, activation) -> Dense(output_dim)
    """

    def __init__(self,
                 units,
                 activation='relu',
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 **kwargs):
        super(FeedForward, self).__init__(**kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.kernel_initializer = initializers.get(kernel_initializer)

    def build(self, input_shape):
        super(FeedForward, self).build(input_shape)

        self.dense_1 = Dense(units=self.units,
                             activation=self.activation,
                             use_bias=self.use_bias,
                             kernel_initializer=self.kernel_initializer)
        self.dense_2 = Dense(units=input_shape[-1],
                             use_bias=self.use_bias,
                             kernel_initializer=self.kernel_initializer)

    def call(self, inputs, **kwargs):
        x = self.dense_1(inputs)
        x = self.dense_2(x)
        return x

    def get_config(self):
        base_config = super(FeedForward, self).get_config()
        config = {
            'units': self.units,
            'activation': self.activation,
            'use_bias': self.use_bias,
            'kernel_initializer': initializers.serialize(self.kernel_initializer)
        }
        return dict(list(base_config.items()) + list(config.items()))


class BiasAdd(Layer):
    def build(self, input_shape):
        super(BiasAdd, self).build(input_shape)
        self.bias = self.add_weight('bias',
                                    shape=(input_shape[-1],),
                                    initializer='zero',
                                    trainable=True)

    def call(self, inputs, **kwargs):
        return K.bias_add(inputs, self.bias)


class Embedding(keras.layers.Embedding):

    def compute_mask(self, inputs, mask=None):
        """为了适配T5，保证第一个token不被mask
        """
        if self._mode == 'embedding':
            mask = super(Embedding, self).compute_mask(inputs, mask)
            if mask is not None:
                mask1 = K.ones_like(mask[:, :1], dtype='bool')
                mask2 = mask[:, 1:]
                return K.concatenate([mask1, mask2], 1)
        else:
            return mask

    def call(self, inputs, mode='embedding'):
        """
        embedding mode: 普通embedding， dense mode: 无bias的Dense，x dot embedding.T
        :param inputs:
        :param mode:
        :return:
        """
        self._mode = mode
        if mode == 'embedding':
            return super(Embedding, self).call(inputs)

        kernel = K.transpose(self.embeddings)
        return K.dot(inputs, kernel)

    def compute_output_shape(self, input_shape):
        if self._mode == 'embedding':
            return super(Embedding, self).compute_output_shape(input_shape)

        return input_shape[:2] + (K.int_shape(self.embeddings)[0],)


class AttentionPooling1D(Layer):
    def __init__(self, h_dim=None, **kwargs):
        super(AttentionPooling1D, self).__init__(**kwargs)
        self.h_dim = h_dim

    def build(self, input_shape):
        super(AttentionPooling1D, self).build(input_shape)
        if not self.h_dim:
            self.h_dim = input_shape[-1]

        self.k_dense = Dense(self.h_dim, use_bias=False, activation='tanh')
        self.o_dense = Dense(1, use_bias=False)

    def call(self, x, mask=None):
        x0 = x
        x = self.k_dense(x0)
        x = self.o_dense(x)
        if mask is not None:
            mask = K.cast(mask, K.floatx())
            mask = K.expand_dims(mask, 2)
            x = x - (1 - mask) * 1e12
        x = K.softmax(x, 1)
        x = K.sum(x0 * x, 1)
        return x

    def compute_output_shape(self, input_shape):
        return (input_shape[0], input_shape[-1])

    def compute_mask(self, inputs, mask=None):
        return None

    def get_config(self):
        base_config = super(AttentionPooling1D, self).get_config()
        config = {'h_dim': self.h_dim}
        return dict(list(base_config.items()) + list(config.items()))


class RelativePositionEmbedding(Layer):
    """相对位置编码
    ref：[Self-Attention with Relative Position Representations](http://arxiv.org/abs/1803.02155）
    """

    def __init__(self, input_dim, output_dim, embedding_initializer='zeros', **kwargs):
        super(RelativePositionEmbedding, self).__init__(**kwargs)
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.embedding_initializer = initializers.get(embedding_initializer)

    def build(self, input_shape):
        super(RelativePositionEmbedding, self).build(input_shape)
        self.embeddings = self.add_weight(name='relativePositionEmbedding',
                                          shape=(self.input_dim, self.output_dim),
                                          initializer=self.embedding_initializer, )

    def call(self, inputs):
        relative_position_idx = self.compute_position_idx(inputs)
        return K.gather(self.embeddings, relative_position_idx)

    def compute_position_idx(self, inputs):
        q, v = inputs
        q_idx = K.arange(0, K.shape(q)[1], dtype='int32')
        q_idx = K.expand_dims(q_idx, 1)
        v_idx = K.arange(0, K.shape(v)[1], dtype='int32')
        v_idx = K.expand_dims(v_idx, 0)
        # 相对位置差
        position_idx = v_idx - q_idx
        max_position = (self.input_dim - 1) // 2
        position_idx = K.clip(position_idx, -max_position, max_position)
        position_idx = position_idx + max_position
        return position_idx

    def compute_output_shape(self, input_shape):
        return (None, None, self.output_dim)

    def compute_mask(self, inputs, mask=None):
        return mask[0]

    def get_config(self):
        base_config = super(RelativePositionEmbedding, self).get_config()
        config = {
            'input_dim': self.input_dim,
            'output_dim': self.output_dim,
            'embedding_initializer': initializers.serialize(self.embedding_initializer)
        }
        return dict(list(base_config.items()) + list(config.items()))


class RelativePositionEmbeddingT5(RelativePositionEmbedding):
    """Google T5的相对位置编码
    来自论文：https://arxiv.org/abs/1910.10683
    """
    def __init__(
        self,
        input_dim,
        output_dim,
        max_distance=128,
        bidirectional=True,
        embeddings_initializer='zeros',
        **kwargs
    ):
        super(RelativePositionEmbeddingT5,
              self).__init__(input_dim, output_dim, **kwargs)
        self.max_distance = max_distance
        self.bidirectional = bidirectional

    def compute_position_ids(self, inputs):
        """T5的相对位置分桶（直接翻译自官方T5源码）
        i-i:   0 1 2 3 4 5 6 7 8 9 10 11 12 13 14...
        f(i-j):0 1 2 3 4 5 6 7 8 8 8  8  9   9  9 ...
        """
        q, v = inputs
        # 计算位置差
        q_idxs = K.arange(0, K.shape(q)[1], dtype='int32')
        q_idxs = K.expand_dims(q_idxs, 1)
        v_idxs = K.arange(0, K.shape(v)[1], dtype='int32')
        v_idxs = K.expand_dims(v_idxs, 0)
        pos_ids = v_idxs - q_idxs
        # 后处理操作
        num_buckets, max_distance = self.input_dim, self.max_distance
        ret = 0
        n = -pos_ids
        if self.bidirectional:
            num_buckets //= 2
            ret += K.cast(K.less(n, 0), 'int32') * num_buckets
            n = K.abs(n)
        else:
            n = K.maximum(n, 0)
        # now n is in the range [0, inf)
        max_exact = num_buckets // 2
        is_small = K.less(n, max_exact)
        val_if_large = max_exact + K.cast(
            K.log(K.cast(n, K.floatx()) / max_exact) /
            np.log(max_distance / max_exact) * (num_buckets - max_exact),
            'int32',
        )
        val_if_large = K.minimum(val_if_large, num_buckets - 1)
        ret += K.switch(is_small, n, val_if_large)
        return ret

    def get_config(self):
        config = {
            'max_distance': self.max_distance,
            'bidirectional': self.bidirectional,
        }
        base_config = super(RelativePositionEmbeddingT5, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class AbsolutePositionEmbeddingTUPE(Layer):
    """
    TUPE式位置与位置交互的位置编码
    出自论文：http://arxiv.org/abs/2006.15595
    """
    def __init__(self,
                 input_dim,
                 output_dim,
                 num_attention_heads,
                 attn_scale_factor=2,
                 relative_position_bias=True,
                 embeddings_initializer='zeros',
                 **kwargs
                 ):
        super(AbsolutePositionEmbeddingTUPE, self).__init__(**kwargs)
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.num_attention_heads = num_attention_heads
        self.attn_scale_factor = attn_scale_factor
        self.relative_position_bias = relative_position_bias
        self.embedding_initializer = initializers.get(embeddings_initializer)
        self.pos_scaling = float(self.output_dim / num_attention_heads * self.attn_scale_factor) ** -0.5

    def build(self, input_shape):
        self.embeddings = self.add_weight(name='absolutePositionEmbedding',
                                          shape=(self.input_dim, self.output_dim),
                                          initializer=self.embedding_initializer, )
        self.q_dense = Dense(self.output_dim, use_bias=False)
        self.k_dense = Dense(self.output_dim, use_bias=False)
        self.pos_ln = LayerNormalization()

    def call(self, inputs, **kwargs):
        q, v = inputs
        seq_len = K.shape(q)[1]
        position_ids = K.arange(0, seq_len + 1, 'int32')  # 增加一个虚拟头结点来解耦cls
        x = K.gather(self.embeddings, position_ids)
        x = self.pos_ln(x)
        q = self.q_dense(x) * self.pos_scaling
        k = self.k_dense(x)
        q = K.reshape(q, (seq_len+1, self.num_attention_heads, -1))
        k = K.reshape(k, (seq_len+1, self.num_attention_heads, -1))

        abs_pos_bias = tf.einsum('jhd,khd->hjk', q, k)
        # p_0 \dot p_0 is  cla to others
        cls_2_other = abs_pos_bias[:, 0, 0]
        # p_1 \dot p_1 is others to cls
        other_2_cls = abs_pos_bias[:, 1, 1]
        # offset
        abs_pos_bias = abs_pos_bias[:, 1:, 1:]
        abs_pos_bias[:, :, 0] = K.reshape(other_2_cls, (-1, 1))
        abs_pos_bias[:, 0, :] = K.reshape(cls_2_other, (-1, 1))
        if self.relative_position_bias:
            rel_pos_bias = self.compute_rel_pos_bias(inputs)
            abs_pos_bias += rel_pos_bias
        return abs_pos_bias

    def compute_rel_pos_bias(self, inputs):
        pass


class DGCNN(Layer):
    """
    膨胀卷积网络，优势是本质是一个CNN，所以速度上比RNNs快，同时通过不同的膨胀系数，如【1，3，5，8】可以来整合全局信息，
    此外，与残差网络结合，解决梯度消失问题，并让信息在多个通道流通。所以在处理序列数据时可以抛弃RNNs而尝试使用该结构。
    ref: https://spaces.ac.cn/archives/5409
    """

    def __init__(self, o_dim=None, k_size=3, dilation_rate=1, skip_connection=True, dropout_rate=None, **kwargs):
        super(DGCNN, self).__init__(**kwargs)

        self.o_dim = o_dim
        self.k_size = k_size
        self.dilation_rate = dilation_rate
        self.skip_connection = skip_connection
        self.dropout_rate = dropout_rate

    def build(self, input_shape):
        super(DGCNN, self).build(input_shape)
        if self.o_dim is None:
            self.o_dim = input_shape[-1]
        self.conv1d = Conv1D(
            self.o_dim * 2,
            self.k_size,
            dilation_rate=self.dilation_rate,
            padding='same',
            name='dgcnn_conv1d'
        )
        if self.skip_connection and self.o_dim != input_shape[-1]:
            self.conv1d_1x1 = Conv1D(self.o_dim, 1)

    def call(self, x, mask=None):
        x0 = x
        if mask is not None:
            mask = K.cast(mask, K.floatx())
            mask = K.expand_dims(mask, 2)
        #         x = x0 * mask if mask is not None else x0
        x0 = Lambda(lambda x_: x_, output_shape=lambda s: s)(x0)  # drop mask so do not put mask to conv1d
        x = self.conv1d(x0)
        x, g = x[:, :, :self.o_dim], x[:, :, self.o_dim:]
        if self.dropout_rate is not None:
            g = K.in_train_phase(K.dropout(g, self.dropout_rate), g)
        g = K.sigmoid(g)
        # mask is none
        mask = mask if mask is not None else K.ones_like(x)

        if self.skip_connection:
            if K.int_shape(x0)[-1] != self.o_dim:
                x0 = self.conv1d_1x1(x0)
            return (x0 * (1 - g) + x * g) * mask
        return x * g * mask

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.o_dim,)

    def compute_mask(self, inputs, mask):
        return mask

    def get_config(self):
        base_config = super(DGCNN, self).get_config()
        config = {
            'o_dim': self.o_dim,
            'k_size': self.k_size,
            'dilation_rate': self.dilation_rate,
            'skip_connection': self.skip_connection,
            'dropout_rate': self.dropout_rate
        }
        return dict(list(base_config.items()) + list(config.items()))


class SinCosPositionEmbedding(Layer):
    def __init__(self, embedding_dim=None, method='add', embedding=None, **kwargs):
        self.method = method  # add or concate
        self.embedding_dim = embedding_dim  # encoder embedding dim, d_pos
        self.embedding = embedding
        super(SinCosPositionEmbedding, self).__init__(**kwargs)

    def call(self, inputs):
        #     PE_2i(p) = sin(p/10000^(2i/d_pos))
        #     PE_2i+1(p) = cos(p/10000^(2i/d_pos))
        batch_size, seq_len, word_emb_dim = K.shape(inputs)[0], K.shape(inputs)[1], K.shape(inputs)[2]
        if not self.embedding_dim or self.method == 'add':
            self.embedding_dim = word_emb_dim
        t = 2 * K.arange(self.embedding_dim / 2, dtype='float32') / K.cast(self.embedding_dim, dtype='float32')
        embedding_wise_pos = 1. / K.pow(10000., t)  # 1/10000 ^(2i/d_pos) , shape = (p_dim/2, )
        embedding_wise_pos = K.expand_dims(embedding_wise_pos, 0)  # (1, p_dim/2)
        word_wise_pos = K.cumsum(K.ones_like(inputs[:, :, 0]), axis=1)  # shape = [batch_size, seq_len]
        word_wise_pos = K.expand_dims(word_wise_pos, 2)  # (batch_size, seq_len, 1)
        position_embedding = K.dot(word_wise_pos, embedding_wise_pos)  # (batch_size, seq_len, p_dim/2)

        position_embedding = K.expand_dims(position_embedding, 3)
        position_embedding = K.reshape(K.concatenate([K.sin(position_embedding), K.cos(position_embedding)], axis=-1),
                                       shape=(batch_size, seq_len, -1))

        if self.method == 'add':
            return inputs + position_embedding

        return K.concatenate([inputs, position_embedding], axis=-1)

    def compute_output_shape(self, input_shape):
        if self.method == 'add':
            return input_shape
        return input_shape[:-1] + (input_shape[-1] + self.embedding_dim,)

    def compute_mask(self, inputs, mask=None):
        return mask


class ConditionalRandomField(Layer):
    """
    带有可调节学习率的CRF，其中调节学习率可以参考[](https://kexue.fm/archives/6418) [](https://kexue.fm/archives/7196)
    sgd优化器下等效放大 lr_multiplier ** 2 , adam下等效放大 lr_multiplier 倍
    """

    def __init__(self, lr_multiplier=1, *args, **kwargs):
        super(ConditionalRandomField, self).__init__(*args, **kwargs)
        self.lr_multiplier = lr_multiplier  # 放大倍数

    def build(self, input_shape):
        super(ConditionalRandomField, self).build(input_shape)
        label_size = input_shape[-1]
        self._trans = self.add_weight(name='crf_trans',
                                      shape=(label_size, label_size),
                                      initializer='glorot_uniform',
                                      trainable=True)
        if self.lr_multiplier != 1:
            K.set_value(self._trans, K.eval(self._trans) / self.lr_multiplier)

    @property
    def trans(self):
        """权重按比例缩放回来"""
        if self.lr_multiplier != 1:
            return self.lr_multiplier * self._trans

        return self._trans

    def compute_mask(self, inputs, mask=None):
        return None

    def call(self, inputs, mask=None):
        # 只是计算loss，并不改变输入
        if mask is not None:
            mask = K.cast(mask, K.floatx())

        return sequence_masking(inputs, mask, 1, 1)

    def path_score(self, inputs, labels):
        """
        :param inputs: (batch_size, timesteps, num_label), obtained from rnn(lstm, bilstm. etc.)
        :param labels: one-hot, (batch_size, timesteps, num_label) , real target series
        :return:  path score
        """
        point_score = tf.einsum('btn,btn->b', inputs, labels)  # 逐标签得分
        trans_score = tf.einsum('bti,ij,btj->b', labels[:, :-1], self.trans, labels[:, 1:])  # 标签转移得分
        return point_score + trans_score

    def log_norm_step(self, inputs, states):
        """递归求解归一化因子"""
        inputs, mask = inputs[:, :-1], inputs[:, -1:]
        states = K.expand_dims(states[0], 2)  # batch_size,output_dim, 1
        trans = K.expand_dims(self.trans, 0)  # 1, output_dim, output_dim
        outputs = K.logsumexp(states + trans, 1)
        outputs += inputs
        outputs = mask * outputs + (1 - mask) * states[:, :, 0]
        return outputs, [outputs]

    def dense_loss(self, y_true, y_pred):
        """y_true需要是one hot形式
        """
        # 导出mask并转换数据类型
        mask = K.all(K.greater(y_pred, -1e6), axis=2, keepdims=True)
        mask = K.cast(mask, K.floatx())
        # 计算目标分数
        y_true, y_pred = y_true * mask, y_pred * mask
        target_score = self.path_score(y_pred, y_true)
        # 递归计算log Z
        init_states = [y_pred[:, 0]]
        y_pred = K.concatenate([y_pred, mask], axis=2)
        input_length = K.int_shape(y_pred[:, 1:])[1]
        log_norm, _, _ = K.rnn(
            self.log_norm_step,
            y_pred[:, 1:],
            init_states,
            input_length=input_length
        )  # 最后一步的log Z向量
        log_norm = K.logsumexp(log_norm, 1)  # logsumexp得标量
        # 计算损失 -log p
        return log_norm - target_score

    def sparse_loss(self, y_true, y_pred):
        """y_true需要是整数形式（非one hot）
        """
        # y_true需要重新明确一下shape和dtype
        y_true = K.reshape(y_true, K.shape(y_pred)[:-1])
        y_true = K.cast(y_true, 'int32')
        # 转为one hot
        y_true = K.one_hot(y_true, K.shape(self.trans)[0])
        return self.dense_loss(y_true, y_pred)

    def dense_accuracy(self, y_true, y_pred):
        """训练过程中显示逐帧准确率的函数，排除了mask的影响
        此处y_true需要是one hot形式
        """
        y_true = K.argmax(y_true, 2)
        return self.sparse_accuracy(y_true, y_pred)

    def sparse_accuracy(self, y_true, y_pred):
        """训练过程中显示逐帧准确率的函数，排除了mask的影响
        此处y_true需要是整数形式（非one hot）
        """
        # 导出mask并转换数据类型
        mask = K.all(K.greater(y_pred, -1e6), axis=2)
        mask = K.cast(mask, K.floatx())
        # y_true需要重新明确一下shape和dtype
        y_true = K.reshape(y_true, K.shape(y_pred)[:-1])
        y_true = K.cast(y_true, 'int32')
        # 逐标签取最大来粗略评测训练效果
        y_pred = K.cast(K.argmax(y_pred, 2), 'int32')
        isequal = K.cast(K.equal(y_true, y_pred), K.floatx())
        return K.sum(isequal * mask) / K.sum(mask)

    def get_config(self):
        config = {
            'lr_multiplier': self.lr_multiplier,
        }
        base_config = super(ConditionalRandomField, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class GatherLastToken(Layer):
    """
    获取最后一个token对应的output，
    如果指定pad_token_id，则是最后一个非pad_token对应的output，否则为序列最后一个token对应的output
    """

    def __init__(self, pad_token_id=None, **kwargs):
        super(GatherLastToken, self).__init__(**kwargs)
        self.pad_token_id = pad_token_id

    def call(self, inputs, **kwargs):
        logits, token_seq = inputs[:2]
        seq_shape = K.shape(token_seq)
        batch_size, seq_length = seq_shape[0], seq_shape[1]
        if self.pad_token_id is None:
            sequence_lengths = [seq_length - 1] * batch_size
        else:
            sequence_lengths = (
                    K.sum(
                        K.cast(
                            K.not_equal(token_seq, self.pad_token_id),
                            dtype='int32',
                        ),
                        -1,
                        keepdims=False,
                    )
                    - 1
            )
        # only tf2
        # return tf.gather(logits, sequence_lengths, batch_dims=1, axis=1)

        indices = K.expand_dims(sequence_lengths, -1)
        return tf.gather_nd(logits, indices, batch_dims=1)

    def compute_output_shape(self, input_shape):
        return (input_shape[0][0], input_shape[0][-1])

    def compute_mask(self, inputs, mask=None):
        return mask[0]


class Loss(Layer):
    """
    特殊层，用于编写灵活loss
    """

    def __init__(self, output_idx=None, **kwargs):
        super(Loss, self).__init__(**kwargs)
        self.output_idx = output_idx  # 输出为inputs的idx部分

    def compute_loss(self, inputs):
        raise NotImplementedError

    def call(self, inputs, **kwargs):
        loss = self.compute_loss(inputs)
        self.add_loss(loss)

        if self.output_idx is None:
            return inputs

        if type(self.output_idx) == list:
            return [inputs[idx] for idx in self.output_idx]
        return inputs[self.output_idx]

    def compute_output_shape(self, input_shape):
        if self.output_idx is None:
            return input_shape
        if type(self.output_idx) == list:
            return [input_shape[idx] for idx in self.output_idx]
        return input_shape[self.output_idx]

    def get_config(self):
        config = {'output_idx': self.output_idx}
        base_config = super(Loss, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class ReWeight(Layer):
    def __init__(self, init_reweight=0, trainable=True, **kwargs):
        super(ReWeight, self).__init__(**kwargs)
        self.init_reweight = init_reweight
        self.trainable = trainable

    def build(self, input_shape):
        initializer = keras.initializers.Constant(self.init_reweight)
        self.beta = self.add_weight(shape=(1,), initializer=initializer, name='reweight', trainable=self.trainable)

    def call(self, x):
        return x * self.beta


custom_objects = {
    'Embedding': Embedding,
    'BiasAdd': BiasAdd,
    'MultiHeadAttention': MultiHeadAttention,
    'LayerNormalization': LayerNormalization,
    'TokenAndPositionEmbedding': TokenAndPositionEmbedding,
    'PositionEmbedding': PositionEmbedding,
    'FeedForward': FeedForward,
    'AttentionPooling1D': AttentionPooling1D,
    'ConditionalRandomField': ConditionalRandomField,
    'DGCNN': DGCNN,
    'SinCosPositionEmbedding': SinCosPositionEmbedding,
    'Loss': Loss,
    'RelativePositionEmbedding': RelativePositionEmbedding,
    'ReWeight': ReWeight,
}

keras.utils.get_custom_objects().update(custom_objects)



# -*- coding: utf-8 -*-
# @Date    : 2020/7/17
# @Author  : mingming.xu
# @Email   : xv44586@gmail.com
# @File    : tokenizers.py
import six
import unicodedata
import json
import re

_open_ = open
is_py2 = six.PY2
if not is_py2:
    basestring = str


def convert_to_str(text, encoding='utf-8', errors='ignore'):
    """字符串转换为str格式（假设输入为utf-8格式）
    """
    if is_py2:
        if isinstance(text, unicode):
            text = text.encode(encoding, errors=errors)
    else:
        if isinstance(text, bytes):
            text = text.decode(encoding, errors=errors)
    return text


def convert_to_unicode(text, encoding='utf-8', errors='ignore'):
    """字符串转换为unicode格式（假设输入为utf-8格式）
    """
    if is_py2:
        if isinstance(text, str):
            text = text.decode(encoding, errors=errors)
    else:
        if isinstance(text, bytes):
            text = text.decode(encoding, errors=errors)
    return text


class open:
    """模仿python自带的open函数，主要是为了同时兼容py2和py3
    """

    def __init__(self, name, mode='r', encoding=None, errors='ignore'):
        if is_py2:
            self.file = _open_(name, mode)
        else:
            self.file = _open_(name, mode, encoding=encoding, errors=errors)
        self.encoding = encoding
        self.errors = errors

    def __iter__(self):
        for l in self.file:
            if self.encoding:
                l = convert_to_unicode(l, self.encoding, self.errors)
            yield l

    def read(self):
        text = self.file.read()
        if self.encoding:
            text = convert_to_unicode(text, self.encoding, self.errors)
        return text

    def write(self, text):
        if self.encoding:
            text = convert_to_str(text, self.encoding, self.errors)
        self.file.write(text)

    def flush(self):
        self.file.flush()

    def close(self):
        self.file.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()


def load_vocab(vocab_path, encoding='utf8', simplified=False, startswith=None):
    """
    从字典文件中读入字典
    """
    token_dict = {}
    with open(vocab_path, encoding=encoding) as reader:
        for line in reader:
            token = line.strip()
            token_dict[token] = len(token_dict)

    if not simplified:
        return token_dict

    new_token_dict = {}
    keep_tokens = []
    if startswith is not None:
        for s in startswith:
            new_token_dict[s] = len(new_token_dict)
            keep_tokens.append(token_dict[s])

    for t, _ in sorted(token_dict.items(), key=lambda kv: kv[1]):
        if t not in new_token_dict:
            keep = True
            if len(t) > 1:
                for c in remove_remark(t):
                    if is_punctuation(c) or is_chinese_char(c):
                        keep = False
                        break
            if keep:
                new_token_dict[t] = len(new_token_dict)
                keep_tokens.append(token_dict[t])

    return new_token_dict, keep_tokens


def save_vocab(save_path, token_dict, encoding='utf8'):
    with open(save_path, 'w', encoding=encoding) as fout:
        for k, _ in sorted(token_dict.items(), key=lambda kv: kv[1]):
            fout.write(k + '\n')


def remove_remark(token):
    # remove any '##' remark at the start of the token
    if token[:2] == '##':
        return token[2:]
    return token


def is_whitespace(char):
    """Checks whether `chars` is a whitespace character."""
    # \t, \n, and \r are technically contorl characters but we treat them
    # as whitespace since they are generally considered as such.
    if char == " " or char == "\t" or char == "\n" or char == "\r":
        return True
    cat = unicodedata.category(char)
    if cat == "Zs":
        return True
    return False


def is_control(char):
    """Checks whether `chars` is a control character."""
    # These are technically control characters but we count them as whitespace
    # characters.
    if char == "\t" or char == "\n" or char == "\r":
        return False
    cat = unicodedata.category(char)
    if cat in ("Cc", "Cf"):
        return True
    return False


def is_punctuation(char):
    """Checks whether `chars` is a punctuation character."""
    cp = ord(char)
    # We treat all non-letter/number ASCII as punctuation.
    # Characters such as "^", "$", and "`" are not in the Unicode
    # Punctuation class but we treat them as punctuation anyways, for
    # consistency.
    if ((cp >= 33 and cp <= 47) or (cp >= 58 and cp <= 64) or
            (cp >= 91 and cp <= 96) or (cp >= 123 and cp <= 126)):
        return True
    cat = unicodedata.category(char)
    if cat.startswith("P"):
        return True
    return False


def _cjk_punctuation():
    return u'\uff02\uff03\uff04\uff05\uff06\uff07\uff08\uff09\uff0a\uff0b\uff0c\uff0d\uff0f\uff1a\uff1b\uff1c\uff1d\uff1e\uff20\uff3b\uff3c\uff3d\uff3e\uff3f\uff40\uff5b\uff5c\uff5d\uff5e\uff5f\uff60\uff62\uff63\uff64\u3000\u3001\u3003\u3008\u3009\u300a\u300b\u300c\u300d\u300e\u300f\u3010\u3011\u3014\u3015\u3016\u3017\u3018\u3019\u301a\u301b\u301c\u301d\u301e\u301f\u3030\u303e\u303f\u2013\u2014\u2018\u2019\u201b\u201c\u201d\u201e\u201f\u2026\u2027\ufe4f\ufe51\ufe54\u00b7\uff01\uff1f\uff61\u3002'


def is_chinese_char(cp):
    """Checks whether CP is the codepoint of a CJK character."""
    # This defines a "chinese character" as anything in the CJK Unicode block:
    #   https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
    #
    # Note that the CJK Unicode block is NOT all Japanese and Korean characters,
    # despite its name. The modern Korean Hangul alphabet is a different block,
    # as is Japanese Hiragana and Katakana. Those alphabets are used to write
    # space-separated words, so they are not treated specially and handled
    # like the all of the other languages.
    cp = ord(cp)
    if ((cp >= 0x4E00 and cp <= 0x9FFF) or  #
            (cp >= 0x3400 and cp <= 0x4DBF) or  #
            (cp >= 0x20000 and cp <= 0x2A6DF) or  #
            (cp >= 0x2A700 and cp <= 0x2B73F) or  #
            (cp >= 0x2B740 and cp <= 0x2B81F) or  #
            (cp >= 0x2B820 and cp <= 0x2CEAF) or
            (cp >= 0xF900 and cp <= 0xFAFF) or  #
            (cp >= 0x2F800 and cp <= 0x2FA1F)):  #
        return True

    return False


class BasicTokenizer(object):
    def __init__(self, start='[CLS]', end='[SEP]'):
        self._token_pad = '[PAD]'
        self._token_unk = '[UNK]'
        self._token_mask = '[MASK]'
        self._token_start = start
        self._token_end = end

    def tokenize(self, text):
        raise NotImplementedError

    def encode(self, first_text, second_text=None, maxlen=None):
        raise NotImplementedError

    def decode(self, tokens):
        raise NotImplementedError

    def truncate_seq(self, first_seq, second_seq=None, maxlen=None, pop_index=-1):
        """按 maxlen 截断两个序列，策略是优先从较长的一个中 pop.(pop_index)"""
        if second_seq is None:
            second_seq = []
        while True:
            total_length = len(first_seq) + len(second_seq)
            if total_length <= maxlen:
                break
            elif len(first_seq) > len(second_seq):
                first_seq.pop(pop_index)
            else:
                second_seq.pop(pop_index)

        return first_seq, second_seq

    def token_to_id(self, token):
        raise NotImplementedError

    def tokens_to_ids(self, tokens):
        return [self.token_to_id(token) for token in tokens]

    def id_to_token(self, id_):
        raise NotImplementedError

    def ids_to_tokens(self, ids):
        return [self.id_to_token(id_) for id_ in ids]


class Tokenizer(BasicTokenizer):
    """Bert tokenizer"""

    def __init__(self, token_dict, do_lower_case=False, *args, **kwargs):
        super(Tokenizer, self).__init__(*args, **kwargs)
        if isinstance(token_dict, six.string_types):
            token_dict = load_vocab(token_dict)
        self._token_dict = token_dict
        self._do_lower_case = do_lower_case
        self._inv_token_dict = {v: k for k, v in token_dict.items()}
        self._vocab_size = len(token_dict)

        # update attribute of particular token
        for token in ['pad', 'unk', 'mask', 'start', 'end']:
            # custom tokens may not have all of this tokens
            try:
                token_id = token_dict[getattr(self, '_token_{}'.format(token))]
                setattr(self, '_token_{}_id'.format(token), token_id)
            except:
                pass

    def _basic_tokenize(self, text):
        text = convert_to_unicode(text)
        if self._do_lower_case:
            text = text.lower()
            """Strips accents from a piece of text."""
            text = unicodedata.normalize("NFD", text)
            text = [cat for cat in text if unicodedata.category(cat) != 'Mn']

        spaced_text = ''
        for char in text:
            if ord(char) == 0 or ord(char) == 0xfffd or is_control(char):  # 过滤非法字符
                continue
            elif is_punctuation(char) or is_chinese_char(char):  # 中文与标点前后加空格切分
                spaced_text += ' ' + char + ' '
            elif is_whitespace(char):  # 空格符直接加一个 ' '
                spaced_text += ' '
            else:
                spaced_text += char
        return spaced_text

    def _word_piece_tokenize(self, word):
        # word piece tokenize
        if word in self._token_dict:
            return [word]

        tokens = []
        start, stop = 0, 0
        while start < len(word):
            stop = len(word)
            while stop > start:
                sub = word[start:stop]
                if start > 0:
                    sub = '##' + sub
                if sub in self._token_dict:
                    break
                stop -= 1
            if start == stop:
                stop += 1
            tokens.append(sub)
            start = stop

        return tokens

    def token_to_id(self, token):
        return self._token_dict.get(token, self._token_unk_id)

    def id_to_token(self, id_):
        return self._inv_token_dict.get(id_, self._token_unk)

    def _tokenize(self, text):
        """token text"""
        spaced_text = self._basic_tokenize(text)
        tokens = []
        for text in spaced_text.strip().split():
            tokens.extend(self._word_piece_tokenize(text))

        return tokens

    def tokenize(self, text, maxlen=None):
        """token text 并按bert格式拼装结果"""
        tokens = self._tokenize(text)
        if self._token_start is not None:
            tokens.insert(0, self._token_start)
        if self._token_end is not None:
            tokens.append(self._token_end)

        if maxlen is None:
            return tokens

        pop_index = int(bool(self._token_end)) + 1
        tokens, _ = self.truncate_seq(tokens, None, maxlen=maxlen, pop_index=-pop_index)
        return tokens

    def encode(self, first_text, second_text=None, maxlen=None, pattern='S*E*E'):
        """
        输出对应的token_ids 与 segment_ids
        :param first_text:
        :param second_text:
        :param maxlen:
        :param pattern: S*E*E / S*ES*E ,区别第二个seq是否有start_token
        :return:
        """
        if isinstance(first_text, basestring):
            first_tokens = self.tokenize(first_text)
        else:
            first_tokens = first_text

        if second_text is None:
            second_tokens = None
        elif isinstance(second_text, basestring):
            if pattern == 'S*E*E':
                start_index = int(bool(self._token_start))
            second_tokens = self.tokenize(second_text)[start_index:]
        else:
            second_tokens = second_text

        if maxlen is not None:
            self.truncate_seq(first_tokens, second_tokens, maxlen, -2)

        first_token_ids = self.tokens_to_ids(first_tokens)
        first_segment_ids = [0] * len(first_token_ids)
        second_token_ids = self.tokens_to_ids(second_tokens) if second_tokens else []
        second_segment_ids = [1] * len(second_token_ids) if second_token_ids else []
        return first_token_ids + second_token_ids, first_segment_ids + second_segment_ids

    def _is_special_token(self, token):
        # 是否是带有 [ ] 的特殊字符
        return bool(token) and token[0] == '[' and token[-1] == ']'

    def decode(self, ids, tokens=None):
        """转换为可读文本"""
        tokens = tokens or self.ids_to_tokens(ids)
        tokens = [token for token in tokens if not self._is_special_token(token)]  # 过滤特殊token
        text, flat = '', False
        for i, token in enumerate(tokens):
            if token[:2] == '##':
                text += token[2:]
            elif len(token) == 1 and is_chinese_char(token):
                text += token
            elif len(token) == 1 and is_punctuation(token):
                text += token + ' '
            elif len(token) == 1 and is_chinese_char(token):
                text += token
            elif i > 0 and is_chinese_char(text[-1]):
                text += token
            else:
                text += ' ' + token
        # format
        text = re.sub(' +', ' ', text)  # 连续空格替换为单个
        text = re.sub('\' (re|m|s|t|ve|d|ll) ', '\'\\1 ', text)  # 右上单引号后接空格 去掉空格
        punctuation = _cjk_punctuation() + '+-/={(<['
        punctuation_regex = '|'.join([re.escape(p) for p in punctuation])
        punctuation_regex = '(%s) ' % punctuation_regex  # 中文符号后的空格删除
        text = re.sub(punctuation_regex, '\\1', text)
        text = re.sub('(\d\.) (\d)', '\\1\\2', text)  # 删除数字小数点中间空格
        return text.strip()

    def rematch(self, text, tokens):
        """原始text 与 tokenize后的token的映射关系"""
        if is_py2:
            text = unicode(text)
        if self._do_lower_case:
            text = text.lower()

        char_mapping = []
        normalized_text = ''
        for i, char in enumerate(text):
            if self._do_lower_case:
                char = unicodedata.normalize("NFD", char)
                char = ''.join([ch for ch in char if unicodedata.category(ch) != 'Mn'])

            char = ''.join([ch for ch in char if  not (ord(ch) == 0 or ord(ch) == 0xfffd or is_control(ch))])
            normalized_text += char
            char_mapping.extend([i]*len(char))

        token_mapping = []
        offset = 0
        for token in tokens:
            if self._is_special_token(token):
                token_mapping.append([])
            else:
                token = remove_remark(token)
                start = normalized_text[offset:].index(token) + offset
                end = start + len(token)
                token_mapping.append(char_mapping[start: end])
                offset = end
        return token_mapping



# -*- coding: utf-8 -*-
# @Date    : 2020/9/27
# @Author  : mingming.xu
# @Email   : xv44586@gmail.com
# @File    : initializers.py

import numpy as np

from toolkit4nlp.backend import keras
from keras.initializers import *


class Sinusoidal(Initializer):
    """
    sinusoidal initializer，主要用于相对位置编码
    ref: [NEZHA: Neural Contextualized Representation for Chinese Language Understanding](http://arxiv.org/abs/1909.00204)
    """

    def __call__(self, shape, dtype=None):
        """
        直接使用Sin-Cos形式的位置编码
        """
        vocab_size, depth = shape
        embeddings = np.zeros(shape)
        for j in range(vocab_size):
            for k in range(depth // 2):
                theta = j / np.power(10000, 2. * k / depth)
                embeddings[j, 2 * k] = np.sin(theta)
                embeddings[j, 2 * k + 1] = np.cos(theta)

        return embeddings


sinusoidal = Sinusoidal

custom_objects = {
    'sinusoidal': sinusoidal,
    'Sinusoidal': Sinusoidal
}
keras.utils.get_custom_objects().update(custom_objects)



#!/usr/bin/env python
# -*- coding: utf8 -*-
# @Date    : 2020/6/29
# @Author   : mingming.xu
# @Email    : xv44586@gmail.com

from setuptools import setup, find_packages

setup(
    name='toolkit4nlp',
    version='0.6.0',
    description='an toolkit for nlp research',
    long_description='toolkit4nlp: https://github.com/xv44586/toolkit4nlp',
    license='Apache License 2.0',
    url='https://github.com/xv44586/toolkit4nlp',
    author='xv44586',
    author_email='xv44586@gmail.com',
    install_requires=['keras<=2.3.1'],
    packages=find_packages()
)
